#ifndef _random_number_h
#define _random_number_h
/**
 * @file random_number.h
 * @author mfh
 * @date 09 Aug 2005
 * 
 * Utility functions for creating random numbers.
 ***************************************************************************/

/**
 * Call this exactly once in your program, before you generate any random 
 * numbers.  You have to specify the seed (as a 32-bit unsigned integer).  
 * Even if you want the same sequence of random numbers each time, you have 
 * to call this function.  Furthermore, this function may not be implemented
 * (in fact, is probably not implemented) using rand() and srand(), so you 
 * shouldn't count on srand() having been called.  But if you are, why are 
 * you using this library anyway?
 *
 * @param seed [IN]  Seed for the random number generator.
 */
void
smvm_init_randomizer (unsigned int seed);

/**
 * Returns a random integer in [low, high].  
 */
int  
smvm_random_integer (int low, int high);


/**
 * Returns a random double in the range [low, high).  This works by first 
 * generating a random double in the range [0,1), and then scaling it to fit.
 * The random number generator used ensures that the double in the range [0,1)
 * has 53 random bits, which is all that a double-precision number can provide.
 *
 * \warn NOTE THE CHANGE IN INTERFACE -- the range is no longer a closed set!!!
 *
 * @param [IN] low
 * @param [IN] high
 */
double
smvm_random_double (double low, double high);


/**
 * This data structure is used to generate random integers out of a range 
 * [low, high] without replacement.  You should consider it an opaque data 
 * structure, and not read from or write to its internals.
 */
struct 
random_integer_from_range_without_replacement_generator_t
{
  int low, high;  
  int* remaining;
  int num_remaining;
};




/**
 * If you want to choose random integers out of the range [low, high] without 
 * replacement, call this function and use the returned object in successive 
 * calls to extract a random integer out of the range without replacement.
 */
struct random_integer_from_range_without_replacement_generator_t*
create_random_integer_from_range_without_replacement_generator (const int low, const int high);

/**
 * Frees a random_integer_from_range_without_replacement_generator_t struct, 
 * once you are done with it.
 */
void
destroy_random_integer_from_range_without_replacement_generator (struct random_integer_from_range_without_replacement_generator_t* gen);

/**
 * Given an initialized random_integer_from_range_without_replacement_generator_t 
 * object, attempts to draw a random integer out of the integers that are remaining.  
 * If an integer remains to choose, it is chosen and assigned to *theint, and zero 
 * is returned.  If no integers remain, nothing is assigned to *theint, and nonzero 
 * is returned.
 */
int
return_random_integer_from_range_without_replacement (int* theint, struct random_integer_from_range_without_replacement_generator_t*);



#endif /* NOT _random_number_h */

