/**************************************************
 * benchmark_subset.c
 * 
 * Determines the "appropriate subset" of the SpMV
 * benchmark to run by estimating the benchmark's
 * run time.
 *
 * author: hormozd
 * last updated: Nov 3, 2005 13:29 PST
 **************************************************/

#include "benchmark_params.h"
#include "benchmark_subset.h"
#include <timer.h>
#include <bcsr_matrix.h>
#include <create_rand.h>
#include <smvm_malloc.h>
#include <smvm_util.h>
#include <stdio.h>
#include <stdlib.h>

/*********************************************************************
 * compute_threshold_dims
 *
 * args:
 *
 * returns: pointer to 2D array of "timer thresholds", i.e. problem
 *          dimensions for each blocksize being tested at which
 *          generating the matrix becomes no longer a "free" operation
 *********************************************************************/
void compute_threshold_dims(hpcc_spmv_params *bench_params, int** tdims) {
  int dim, r, c, r_curr, c_curr, bw;
  double time1, time2;
  struct bcsr_matrix_t *A = NULL;

  WITH_DEBUG2(fprintf(stderr, "=== compute_threshold_dims ===\n"));

  /* fill tdims array by running SpMV timings */
  for (r = 0; r < bench_params->n_row_blockdims; r++) {
    for (c = 0; c < bench_params->n_col_blockdims; c++) {
      for (dim = bench_params->mindim; dim <= bench_params->maxdim; dim *= 2) {
        r_curr = bench_params->row_blockdims[r];
        c_curr = bench_params->col_blockdims[c];
        WITH_DEBUG2(fprintf(stderr, "\tr,c,dim = %d,%d,%d\n", r_curr, c_curr, dim));
        bw = MAX(dim/r_curr,dim/c_curr);

        /* create matrix and time generator */
        time1 = get_seconds();
        A = create_random_matrix_banded_by_nnz_per_row(dim/r_curr, dim/c_curr,
              r_curr, c_curr, bench_params->nnz_per_row_max/c_curr, bw, bw);
        time2 = get_seconds();
        destroy_bcsr_matrix (A);
        A = NULL;

        if (time2-time1 >= .1) {
          tdims[r][c] = dim/2;
          break;
        }
      }
    }
  }
}

/***********************************************************************
 * max_problem_dimension
 *
 * args: benchmark parameters
 *       threshold_dims - 2D array of "timer thresholds"
 *                        corresponding to different blocksizes
 *
 * returns: maximum problem dimension we can run
 ***********************************************************************/
int max_problem_dimension(hpcc_spmv_params *bench_params, int **threshold_dims)
{
  int dim, logdim, r, c, allowed_time, bw;
  int r_curr, c_curr, nnz_per_row_tested, nnz_curr;
  double ***accum_time_ests;
  double time1, time2, est_run_time = 0, time_est_prev, run_time_base;
  struct bcsr_matrix_t *A = NULL;

  allowed_time = 60*bench_params->tmax;
  nnz_per_row_tested = bench_params->nnz_per_row_max - 
                       bench_params->nnz_per_row_min + 1;

  /* allocate space to hold accumulated estimated running time for each
     problem dimension that could be tested */
  accum_time_ests = smvm_calloc(bench_params->numdims, sizeof(double**));

  for (logdim = 0; logdim < bench_params->numdims; logdim++) {
    accum_time_ests[logdim] = smvm_calloc(bench_params->n_row_blockdims, sizeof(double*));

    for (r = 0; r < bench_params->n_row_blockdims; r++) {
      accum_time_ests[logdim][r] = smvm_calloc(bench_params->n_col_blockdims, sizeof(double));
    }
  }

  /* fill in array that estimates benchmark runtime for each problem
     dimension. also use this to estimate runtime for running every
     trial in our test space */
  for (r = 0; r < bench_params->n_row_blockdims; r++) {
    for (c = 0; c < bench_params->n_col_blockdims; c++) {
      logdim = 0;

      for (dim = bench_params->mindim; dim < threshold_dims[r][c]; dim *= 2) {
        logdim++;
      }

      r_curr = bench_params->row_blockdims[r];
      c_curr = bench_params->col_blockdims[c]; 

      /* run matrix generator once and time it */
      bw = MAX(dim/r_curr,dim/c_curr);
      time1 = get_seconds();
      A = create_random_matrix_banded_by_nnz_per_row(dim/r_curr, dim/c_curr,
            r_curr, c_curr, bench_params->nnz_per_row_max/c_curr, bw, bw);
      time2 = get_seconds();
      destroy_bcsr_matrix (A);
      A = NULL;
      accum_time_ests[logdim][r][c] = time2 - time1;
      est_run_time += nnz_per_row_tested*(time2 - time1);
      logdim++;

      /* now compute estimate of benchmark runtime for one nnz/row value
         for this particular blocksize by doubling this runtime until
         we reach maxdim */
      for (dim *= 2; dim <= bench_params->maxdim; dim *= 2) {
        accum_time_ests[logdim][r][c] = 2*accum_time_ests[logdim-1][r][c];
        est_run_time += nnz_per_row_tested*accum_time_ests[logdim][r][c];
        logdim++;
      }
    }
  }

  /* now compute largest problem dimension we can go to. first step is
     to cut down on the amount of nnz/row trials. if that doesn't work,
     then cut the dimension and try again */
  logdim--;
  dim = bench_params->maxdim;
  time_est_prev = est_run_time;

  while (dim >= bench_params->mindim && est_run_time >= 1.1*allowed_time) {
    run_time_base = time_est_prev;

    /* if we've gone through the loop before, cut off the largest
       problem dimension from the test space */
    if (dim < bench_params->maxdim) {
      for (r = 0; r < bench_params->n_row_blockdims; r++) {
        for (c = 0; c < bench_params->n_col_blockdims; c++) {
          run_time_base -= nnz_per_row_tested*accum_time_ests[logdim+1][r][c];
        }
      }
    }

    time_est_prev = run_time_base;

    /* are we within alotted time? if not, cut down on nnz/row */
    nnz_curr = nnz_per_row_tested;

    while (est_run_time > 1.1*allowed_time && nnz_curr > 1) {
      nnz_curr--;
      est_run_time = run_time_base*nnz_curr/nnz_per_row_tested;
    }

    logdim--;
    dim = dim/2;
  }

  /* collect final results */
  if (dim < bench_params->maxdim) {
    dim *= 2;
    logdim += 2;
  }

  bench_params->maxdim = dim;
  bench_params->numdims = logdim;
  bench_params->est_runtime = est_run_time/60 + 1;
  bench_params->nnz_per_row_to_run = nnz_curr;

printf("nnz/row to test = %d\n", nnz_curr);

  /* cleanup */
  for (logdim = 0; logdim < bench_params->numdims; logdim++) {
    for (r = 0; r < bench_params->n_row_blockdims; r++) {
      smvm_free(accum_time_ests[logdim][r]);
    }

    smvm_free(accum_time_ests[logdim]);
  }

  smvm_free(accum_time_ests);
  return dim;
}
