/**********************************************************
 * hpcc_spmv_benchmark.c
 *
 * Front-end for the SpMV Benchmark run HPC Challenge style
 *
 * author: hormozd
 * last updated: Nov 2 2005, 20:43 PST
 **********************************************************/

#include "benchmark_params.h"
#include "benchmark_subset.h"
#include <smvm_benchmark.h>
#include <smvm_timing_results.h>
#include <smvm_malloc.h>
#include <smvm_util.h>
#include <math.h>
#include <timer.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/****************************
 * median
 * 
 * returns median of an array
 ****************************/
double median(double *a, int nelems) {
  /* sort input array */
  qsort(a, nelems, sizeof(double), smvm_compare_double);

  /* what we do depends on whether nelems is even or odd */
  if (nelems % 2 == 1)
    return a[nelems/2];
  else
    return (a[nelems/2]+a[nelems/2 + 1])/2;
}

/*******************************************************
 * round_to_int
 *
 * rounds a floating-point number to the nearest integer
 * only works for positive arguments
 *******************************************************/
double round_to_int(double x) {
  return floor(x+0.5);
}

/*****************************************
 * pick_nnz_row_vals
 *
 * selects which nnz/row trials to conduct
 *****************************************/
void pick_nnz_row_vals(int *vals, int nnz_row_run, int nnz_row_min, int nnz_row_max) {
  int idx;
  double spacing, nnz_to_test;

  if (nnz_row_run == 1)
    vals[0] = (nnz_row_min + nnz_row_max)/2;
  else if (nnz_row_run >= 2) {
    vals[0] = nnz_row_min;
    vals[nnz_row_run-1] = nnz_row_max;

    if (nnz_row_run > 2) {
      spacing = (double)(nnz_row_max - nnz_row_min)/(nnz_row_run-1);
      nnz_to_test = nnz_row_min + spacing;

      for (idx = 1; idx <= nnz_row_run-2; idx ++) {
        vals[idx] = (int)round_to_int(nnz_to_test);
        nnz_to_test += spacing;
      }
    }
  }
}

/******************************************************
 * blockdim_idx
 *
 * find index for mflop data for a particular blocksize
 * assumes 1x1 starts the data array, and the blocksize
 * referenced by r_idx = 0, c_idx = 0 is 1x1
 ******************************************************/
int blockdim_idx(int r_idx, int c_idx, int numdims, int nnz_vals, int n_rdims) {
  return r_idx*n_rdims*numdims*nnz_vals + c_idx*numdims*nnz_vals;
}

/********************************************************
 * mflop_idx
 *
 * find index for mflop data based on blocksize, dim, and
 * nnz/row value
 ********************************************************/
int mflop_idx(int r_idx, int c_idx, int nnz_idx, int dim_idx, int numdims, int nnz_vals, int n_rdims) {
  return blockdim_idx(r_idx,c_idx,numdims,nnz_vals,n_rdims)+
         nnz_idx*numdims + dim_idx;
}

/*************************************************************
 * hpcc_spmv_benchmark
 *
 * driver function for benchmark
 * arguments: benchmark parameters
 * returns: error code; "true" output is printed on the screen
 *************************************************************/
int hpcc_spmv_benchmark(hpcc_spmv_params* bench_params) {
  long int maxdim_intermediate;
  int dim, r, c, dimcheck, num_unblocked_runs, max_blocked_runs;
  int num_runs = 0, nnz_row, r_curr, c_curr;
  int len, mflop_max_idx = 0, nnz_idx, logdim, bigidx;
  int lookahead_idx, nnz_per_row_med, next_nz_row;
  int **threshold_dims;
  int *r_at_max_mflop, *c_at_max_mflop, *nnz_row_vals;
  double time1, time2, unblocked_mflop_max = 0, blocked_mflop_max = 0;
  double mflop_rate, unblocked_mflop_median, blocked_mflop_median;
  double slope;
  double *mflop_rates;
  char *outfilename;
  struct SMVM_timing_results timing_results;
  struct SMVM_parameters spmv_params;
  FILE *outfp;

  if (smvm_debug_level() > 0)  {
    fprintf (stderr, "=== hpcc_smvm_benchmark ===\n");
  }

  /* determine the absolute largest problem dimension allowed
   * us by the memory restriction */

  /* BUG NOTE (hormozd, 23:10 PST, 11/5/2005): for some reason, this
     calculation overflows, even though I'm using a long. the end result
     for, to give an example, memmax = 3000 should fit in an int, even
     though the numerator would overflow in an int but not a long (but 
     shouldn't using a long to handle this numerator solve it?) */

#if 0
  maxdim_intermediate = 1048576*((long)bench_params->memmax);
  maxdim_intermediate = maxdim_intermediate/
                           (12*bench_params->nnz_per_row_max+20);
#endif
  /* BUG FIX (mhoemmen, 10:20 PST, 11/6/2005): the usual way to fix 
   * int overflow bugs in intermediate calculations is to perform the
   * calculations using double-precision floating point (with its much
   * larger exponent range) and then cast the results back to int. */
  {
    double mi = 1048576.0 * (double) (bench_params->memmax);
    mi = mi / (12.0 * (double) (bench_params->nnz_per_row_max) + 20.0);
    maxdim_intermediate = mi;
  }

  bench_params->maxdim = maxdim_intermediate;

  WITH_DEBUG(fprintf(stderr, "absolute maxdim = %d\n", bench_params->maxdim));

  /* now find largest problem dimension we can test given that we're
     doing problem doubling */
  bench_params->numdims = 0;

  for (dim = bench_params->mindim; dim <= bench_params->maxdim; dim *= 2) {
    bench_params->numdims++;
  }

  bench_params->maxdim = dim/2;
  WITH_DEBUG(fprintf(stderr, "practical maxdim = %d\n", bench_params->maxdim));

  /* if the number of dimension sizes that we're going to test is zero, we
     have a problem... */
  if (bench_params->numdims == 0)
    return -1;

  /* allocate space for threshold_dims array */
  threshold_dims = smvm_calloc(bench_params->n_row_blockdims,sizeof(int*));

  for (r = 0; r < bench_params->n_row_blockdims; r++) {
    threshold_dims[r] = smvm_calloc(bench_params->n_col_blockdims,sizeof(int));
  }

  /* these steps are only needed if we're sticking to the time limit */
  if (bench_params->runall == 0) {
    /* compute "threshold dimensions", for which the estimated runtime of 
       the benchmark is on the order of seconds */
    compute_threshold_dims (bench_params, threshold_dims);

    /* now find maximum allowed problem dimension */
    dimcheck = max_problem_dimension (bench_params, threshold_dims);

    /* check return value of max problem dimension */
    printf("max problem dimension = %d\n", dimcheck);
    printf("estimated runtime = %d minutes\n", bench_params->est_runtime);

    if (dimcheck < 0) {
      smvm_free(threshold_dims);
      return dimcheck;
    }
  }
  else {
    bench_params->nnz_per_row_to_run = bench_params->nnz_per_row_max -
                                       bench_params->nnz_per_row_min + 1;
  }

  /* allocate array that contains nnz/row values to run and fill it */
  nnz_per_row_med = (bench_params->nnz_per_row_min + bench_params->nnz_per_row_max)/2;
  nnz_row_vals = smvm_malloc(bench_params->nnz_per_row_to_run*sizeof(int));
  pick_nnz_row_vals(nnz_row_vals,bench_params->nnz_per_row_to_run,bench_params->nnz_per_row_min,bench_params->nnz_per_row_max);

  /* now set up benchmark */
  num_unblocked_runs = bench_params->numdims*(bench_params->nnz_per_row_max-
                          bench_params->nnz_per_row_min+1);
  max_blocked_runs = (bench_params->n_row_blockdims*bench_params->n_col_blockdims-1)*
                        num_unblocked_runs;
  mflop_rates = smvm_calloc(num_unblocked_runs+max_blocked_runs,sizeof(double));
  outfilename = smvm_calloc(strlen(bench_params->outfilebase)+7+
                   bench_params->rmaxlen+bench_params->cmaxlen,sizeof(char));
  r_at_max_mflop = smvm_calloc(bench_params->n_row_blockdims*
                      bench_params->n_col_blockdims,sizeof(int)+1);
  c_at_max_mflop = smvm_calloc(bench_params->n_row_blockdims*
                      bench_params->n_col_blockdims,sizeof(int)+1);

  time1 = get_seconds();

  for (r = 0; r < bench_params->n_row_blockdims; r++) {
    for (c = 0; c < bench_params->n_col_blockdims; c++) {
      r_curr = bench_params->row_blockdims[r];
      c_curr = bench_params->col_blockdims[c];
      nnz_idx = 0;

      for (nnz_row = bench_params->nnz_per_row_min; nnz_row <= bench_params->nnz_per_row_max; nnz_row++) {
        for (dim = bench_params->mindim; dim <= bench_params->maxdim; dim *= 2) {
          /* only run benchmark trial if
               (1) problem size is small enough, or
               (2) we're doing an appropriate value of nnz/row */
          if (dim < threshold_dims[r][c] || nnz_row == nnz_row_vals[nnz_idx]) {
            /* set up benchmark run */
            spmv_params.m = dim;
            spmv_params.n = dim;
            spmv_params.r = r_curr;
            spmv_params.c = c_curr;
            spmv_params.num_trials = 10;
            spmv_params.percent_fill = -nnz_row;

            spmv_params.dataoutfile = fopen(bench_params->rawfile,"w");

            smvm_benchmark_with_results(&spmv_params, &timing_results);

            fclose(spmv_params.dataoutfile);
            mflop_rate = timing_results.mflops/timing_results.t_median;

            /* collect data */
            mflop_rates[num_runs] = mflop_rate;

            if (r == 0 && c == 0 && mflop_rate > unblocked_mflop_max)
              unblocked_mflop_max = mflop_rate;
            else if (r != 0 || c != 0) {
              if (mflop_rate > blocked_mflop_max) {
                blocked_mflop_max = mflop_rate;
                r_at_max_mflop[0] = r_curr;
	        c_at_max_mflop[0] = c_curr;
	        r_at_max_mflop[1] = 0;
	        c_at_max_mflop[1] = 0;
	        mflop_max_idx = 1;
	      }
	      else if (mflop_rate == blocked_mflop_max) {
		r_at_max_mflop[mflop_max_idx] = r_curr;
		c_at_max_mflop[mflop_max_idx] = c_curr;
		mflop_max_idx++;
	      }
            }
          }

          num_runs++;
        }

        if (nnz_row == nnz_row_vals[nnz_idx])
          nnz_idx++;
      }
    }
  }

  time2 = get_seconds();
  bench_params->actual_runtime = time2 - time1;

  /* fill in tests that weren't done using interpolation and write
     resulting data to output file */
  bigidx = 0;

  for (r = 0; r < bench_params->n_row_blockdims; r++) {
    for (c = 0; c < bench_params->n_col_blockdims; c++) {
      r_curr = bench_params->row_blockdims[r];
      c_curr = bench_params->col_blockdims[c];

      len = sprintf(outfilename,"%s_%dx%d.csv",
                    bench_params->outfilebase, r_curr, c_curr);
      outfilename[len] = '\0';

      /* open output file for mflop rate info */
      outfp = fopen(outfilename,"w");

      if (outfp == NULL) {
        printf("ERROR: unable to open output file for %d x %d case!", r_curr, c_curr);
        /* free what needs to be freed here */
        return -1;
      }

      /* write top row of output file */
      for (dim=bench_params->mindim;dim<=bench_params->maxdim;dim*=2) {
        fprintf(outfp,",%d",dim);
      }
      fprintf(outfp,"\n");

      for (nnz_row = bench_params->nnz_per_row_min; nnz_row <= bench_params->nnz_per_row_max; nnz_row++) {
        /* write current nnz/row value to output file */
        fprintf(outfp,"%d",nnz_row);

        for (logdim = 0; logdim < bench_params->numdims; logdim++) {
          /* is interpolation needed to fill this data point? */
          if (mflop_rates[bigidx] == 0) {

            if (bench_params->nnz_per_row_to_run == 1) {
              lookahead_idx = bigidx + bench_params->numdims*(nnz_per_row_med - nnz_row);
              mflop_rates[bigidx] = mflop_rates[lookahead_idx];
            }
            else {
              lookahead_idx = bigidx+bench_params->numdims;
              next_nz_row = nnz_row+1;

              while (mflop_rates[lookahead_idx] == 0) {
                lookahead_idx += bench_params->numdims;
                next_nz_row++;
              }

              mflop_rates[bigidx] = mflop_rates[lookahead_idx];

              /* compute slope for linear interpolation. note that bigidx-numdims
                 will always be filled. */
              slope = (mflop_rates[lookahead_idx] - mflop_rates[bigidx-bench_params->numdims])/(next_nz_row-nnz_row+1);

              /* now fill in value */
              mflop_rates[bigidx] = mflop_rates[bigidx-bench_params->numdims]+slope;
            }
          }

          /* write data to output file */
          fprintf(outfp,",%f",mflop_rates[bigidx]);

          bigidx++;
        }

        fprintf(outfp,"\n");
      }

      fclose(outfp);
    }
  }

  /* later on: organize data to make "comprehensive" graph */

  /* get median mflop numbers for both unblocked and blocked cases
     to complete data analysis (max numbers were already collected
     during the benchmark runs) */
  unblocked_mflop_median = median(mflop_rates, num_unblocked_runs);
  blocked_mflop_median = median(mflop_rates + num_unblocked_runs, max_blocked_runs);

  /* output the magic four numbers */
  printf("\t|Max\t\t|Median\n");
  printf("-------\t|---\t\t|------\n");
  printf("1x1\t|%f\t|%f\n", unblocked_mflop_max, unblocked_mflop_median);
  printf("Blocked\t|%f\t|%f\n", blocked_mflop_max, blocked_mflop_median);
  printf("\nblocked max occurs for blocksizes\n");

  mflop_max_idx = 0;
  while (r_at_max_mflop[mflop_max_idx] != 0
         && c_at_max_mflop[mflop_max_idx] != 0)
  {
    printf("\t%dx%d\n",r_at_max_mflop[mflop_max_idx],c_at_max_mflop[mflop_max_idx]);
    mflop_max_idx++;
  }

  printf("actual runtime: %d minutes\n", (int)((time2-time1)/60));

  /* cleanup */
  for (r = 0; r < bench_params->n_row_blockdims; r++)
    smvm_free(threshold_dims[r]);

  smvm_free(threshold_dims);
  smvm_free(mflop_rates);
  smvm_free(r_at_max_mflop);
  smvm_free(c_at_max_mflop);
  smvm_free(outfilename);
  smvm_free(nnz_row_vals);
  return 0;
}

/****************************************
 * main
 *
 * function: runs the benchmark (duh!)
 * arguments: time and memory constraints
 ****************************************/
int main(int argc, char **argv) {
  int benchcheck, t_max, mem_max;
  int rowdims[] = {1,2,3,4,6,8}, coldims[] = {1,2,3,4,6,8};
  hpcc_spmv_params bench_params;

  smvm_set_debug_level_from_environment ();

  /* obtain arguments. should be invoked like this from
     the command line:

       hpcc_benchmark [tmax] [memmax]

     where tmax is the maximum allowed time in minutes, and memmmax is
     the memory constraint in megabytes */
  if (argc < 3) {
    printf("\nIncorrect number of arguments. Usage is\n");
    printf("\n\thpcc_benchmark [tmax] [memmax] (runall)\n");
    printf("\nwhere tmax is the time limit in minutes and\n");
    printf("memmax is the memory limit in megabytes.\n");
    printf("runall is an optional parameter that if nonzero\n");
    printf("says to run the benchmark with no time limit\n");
    printf("(default value is zero)\n\n");
    return 1;
  }

  /* process arguments and check them */
  t_max = atoi(argv[1]);
  mem_max = atoi(argv[2]);

  if (argc >= 4)
    bench_params.runall = atoi(argv[3]);
  else
    bench_params.runall = 0;

  if (t_max <= 0 || mem_max <= 0) {
    printf("Error: Both tmax and memmax must be positive!\n");
    return 2;
  }

  /* fill benchmark parameter struct */
  bench_params.row_blockdims = rowdims;
  bench_params.col_blockdims = coldims;
  bench_params.n_row_blockdims = 6;
  bench_params.n_col_blockdims = 6;
  bench_params.nnz_per_row_min = 50;
  bench_params.nnz_per_row_max = 60;
  bench_params.mindim = 512;
  bench_params.outfilebase = "mflops";
  bench_params.rawfile = "benchmarkdebug.csv";
  bench_params.tmax = t_max;
  bench_params.memmax = mem_max;
  bench_params.rmaxlen = 1;
  bench_params.cmaxlen = 1;

  /* call "driver routine" */
  benchcheck = hpcc_spmv_benchmark(&bench_params);

  /* we're done! return value returned by driver routine */
  return benchcheck;
}
