/**
 * @file prompt.y
 * @author mfh
 * @since 05 Jul 2005
 * @date 05 Jul 2005
 *
 * Simple shell for testing sparse matrix conversion.  It accepts the following 
 * commands:
 *
 * load <filename> <file-format>;
 *   Assumes that the given sparse matrix file <filename> is in the given 
 *   format <file-format>, and loads the file into the current sparse matrix.
 *   The shell only works with one sparse matrix at a time; it is an implicit
 *   argument of all the shell commands.  Supported file formats are:  
 *   `HARWELL_BOEING', `MATRIX_MARKET' and `MATLAB'.
 *
 * save <filename> <file-format>;
 *   Save the current sparse matrix to the given filename in the given format.
 *   Supported file formats are:  `HARWELL_BOEING', `MATRIX_MARKET' and `MATLAB'.
 *
 * format?
 *   Prints the internal storage format of the current sparse matrix.
 *
 * convert <internal-format>;
 *   Converts the current sparse matrix from its current internal format to a 
 *   different internal format.  Supported internal formats are:  `CSC', `CSR',
 *   `COO', `BCOO', `BCSR', `JAD'.
 *
 * exit;
 *   Frees the current sparse matrix and exits the shell.
 */

%{
#include "config.h"
#include "sparse_matrix.h"
#include "sparse_matrix_ops.h"

#include <get_options.h>
#include <smvm_malloc.h>
#include <smvm_util.h>
#include <timer.h>

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

struct sparse_matrix_t* A = NULL;

int
yylex (void);

void
yyerror (char const * s)
{
  fprintf (stderr, "*** ERROR: %s ***\n", s);
}

double seconds = 0.0;
%}

/* Bison declarations */

%union {
  char* str;  /* For reading strings */
}


/* Terminal symbols */
%token <str> STRING 
%token LOAD SAVE FORMAT CONVERT EXIT

%destructor { fprintf (stderr, "Calling destructor on symbol %s\n", $$); free ($$); } STRING 


%%

input:  /* empty */
	| input line          { printf ("$ "); }

line:   '\n'                  
	| stmtlist '\n'   

stmtlist:  /* empty */
	| stmtlist stmt 
	;

/* filename first, then format */
stmt:   LOAD STRING STRING ';' { if (A) destroy_sparse_matrix (A);
                               seconds = get_seconds ();
                               A = load_sparse_matrix (sparse_matrix_file_format_string_to_enum ($3), $2);
			       seconds = get_seconds () - seconds;
			       printf ("Loading the sparse matrix took %g seconds.\n", seconds);
			     }
/* filename first, then format */
	| SAVE STRING STRING ';' { seconds = get_seconds ();
	                         save_sparse_matrix ($2, A, sparse_matrix_file_format_string_to_enum ($3));
			         seconds = get_seconds () - seconds;
				 printf ("Saving the sparse matrix took %g seconds.\n", seconds);
			       }
	| FORMAT               { if (A == NULL)
	                           printf ("Sparse matrix is NULL!\n");
			         else
				   printf ("%s\n", sparse_matrix_format_string (A));
			       }
	| CONVERT STRING ';'   { seconds = get_seconds ();
	                         sparse_matrix_convert (A, sparse_matrix_storage_format_string_to_enum ($2));
				 seconds = get_seconds () - seconds;
				 printf ("Converting the sparse matrix took %g seconds.\n", seconds);
                               }
	| EXIT ';'             { if (A) destroy_sparse_matrix (A);
	                         printf ("Goodbye!\n");
				 exit (EXIT_SUCCESS);
			       }
	;

%%


int
main (int argc, char** argv)
{
  return yyparse ();
}

