/**
 * @file block_size_parser.c
 * @author mfh
 * @since 2004 Nov 28
 * @date 2004 Nov 29
 *
 * Functions for parsing a list of triples:  r,c,frequency in CSV format.  
 * These represent optimal block dimensions and the number of times that 
 * pair of dimensions appears.
 */
#include "config.h"
#include "block_size_parser.h"
#include <smvm_malloc.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>


/**
 * Reads a single line from the block size data file, parses it and extracts
 * the (r,c,frequency) triplet.
 */
static int
parse_block_sizes_line (struct block_size_data_t *data, FILE *infile, 
			const int datum_number)
{
  const int MAX_LINE_LENGTH = 200;
  char      line[201];
  char     *token = NULL;
  int       r, c, frequency;

  fgets (line, MAX_LINE_LENGTH, infile);
  /* fprintf (stderr, "%s\n", line); */
  
  if (ferror (infile))
    {
      fprintf (stderr, "*** parse_block_sizes_line: failed to read datum "
	       "%d from block sizes file ***\n", datum_number);
      return -1;
    }

  /* Extract r token */
  token = strtok (line, ",");
  if (token == NULL)
    {
      /*fprintf (stderr, "*** parse_block_sizes_line: at datum %d, failed "
	       "to extract first token ***\n", datum_number);
	       */
      return -2;
    }
  /* token points to r; convert to int */
  r = strtol (token, NULL, 10);
  if (errno)
    {
      fprintf (stderr, "*** parse_block_sizes_line: at datum %d, failed "
	       "to convert r = %s to int ***\n", datum_number, token);
      return -3;
    }

  /* Extract c token */
  token = strtok (NULL, ",");
  if (token == NULL)
    {
      /*
      fprintf (stderr, "*** parse_block_sizes_line: at datum %d, failed "
	       "to extract second token ***\n", datum_number);
	       */
      return -2;
    }
  /* token points to c; convert to int */
  c = strtol (token, NULL, 10);
  if (errno)
    {
      fprintf (stderr, "*** parse_block_sizes_line: at datum %d, failed "
	       "to convert c = %s to int ***\n", datum_number, token);
      return -3;
    }

  /* Extract frequency token */
  token = strtok (NULL, ",");
  if (token == NULL)
    {
      /*
      fprintf (stderr, "*** parse_block_sizes_line: at datum %d, failed "
	       "to extract third token ***\n", datum_number);
       */
      return -2;
    }
  /* token points to frequency; convert to int */
  frequency = strtol (token, NULL, 10);
  if (errno)
    {
      fprintf (stderr, "*** parse_block_sizes_line: at datum %d, failed "
	       "to convert frequency = %s to int ***\n", datum_number, token);
      return -3;
    }

  data->r = r;
  data->c = c;
  data->frequency = frequency;

  return 0;
}


/**
 * Appends to the given array of block_size_data_t.
 *
 * @param dataarray [IN/OUT]  Pointer to array of block_size_data_t.
 * @param curlen [IN/OUT]     Pointer to current number of elements in array.
 *                            On exit, incremented by one for the new element.
 * @param maxlen [IN/OUT]     Pointer to current number of allocated spots in
 *                            the array, which is no less than the current
 *                            number of elements in the array.  May be 
 *                            increased on exit.
 */
static void
append_to_bsdt_array (struct block_size_data_t **dataarray, int *curlen, 
		      int *maxlen, const struct block_size_data_t *datum)
{
  if (*curlen >= *maxlen)
    *maxlen = 2 * (*maxlen);

  *dataarray = (struct block_size_data_t*) 
    smvm_realloc (*dataarray, *maxlen * sizeof (struct block_size_data_t));
  (*dataarray)[*curlen] = *datum;
  (*curlen)++;
}


int
parse_block_sizes_file (struct block_size_data_t **dataarray, 
			int *dataarraylen, const char *filename)
{
  FILE *infile;
  int data_count = 0;
  int n = 0;
  int maxlen = 4;
  char line[201];
  const int MAX_LINE_LENGTH = 200;

  *dataarray = (struct block_size_data_t*) 
    smvm_malloc (maxlen * sizeof (struct block_size_data_t));
  *dataarraylen = 0;
  
  infile = fopen (filename, "r");
  if (infile == NULL)
    {
      fprintf (stderr, "*** parse_block_sizes_file: failed to open block "
	       "sizes file %s ***\n", filename);
      return -1;
    }

  /* The first line contains only labels */
  fgets (line, MAX_LINE_LENGTH, infile);
  if (ferror (infile))
    {
      fprintf (stderr, "*** parse_block_sizes_file: failed to read labels "
	       "from block sizes file %s ***\n", filename);
      fclose (infile);
      return -1;
    }

  /* Now go through the file and collect the r,c,frequency triplets */
  while (! feof (infile))
    {
      struct block_size_data_t datum;
      int retcode;

      retcode = parse_block_sizes_line (&datum, infile, data_count);
      if (retcode == 0)
	append_to_bsdt_array (dataarray, &data_count, &maxlen, &datum);
      else
	{
	  /* fprintf (stderr, "*** parse_block_sizes_file: failure at reading "
		   "datum %d from file %s ***\n", data_count, filename);
          */
	  break; /* We have some data, so keep it */
	}
    }
  *dataarraylen = data_count;

  if (data_count == 0)
    {
      fprintf (stderr, "*** parse_block_sizes_file: file %s contains no "
	       "data! ***\n", filename);
      fclose (infile);
      return -4;
    }

  if (0 != fclose (infile))
    {
      fprintf (stderr, "*** parse_block_sizes_file: failed to close block "
	       "sizes file %s ***\n", filename);
      return -1;
    }
  return 0;
}

