#ifndef _hpcc_h
#define _hpcc_h

#ifdef SPMV_STANDALONE
#else
#  include <hpcc.h>
#endif /* NOT SPMV_STANDALONE */

struct SMVM_benchmark_results;


/**
 * A proxy for HPCC_Params.  We want to be able to build and test the SpMV 
 * benchmark without needing HPCC itself to build and run (that requires 
 * installing MPICH and the BLAS).  So all accesses to HPCC_Params pass 
 * through this proxy.  If we are using the SpMV benchmark in standalone
 * mode, the HPCC_Params object is not used; instead, other parameters 
 * stand in for the needed parameters in HPCC_Params.
 */
struct my_HPCC_Params;

#ifdef SPMV_STANDALONE
/**
 * When we are running the SpMV benchmark in standalone mode, this function 
 * allocates and initializes a my_HPCC_Params object, and returns a pointer 
 * to it.
 *
 * @param maxmem [IN]         Maximum memory in bytes that the benchmark may use.
 * @param fill [IN]           Fill fraction of the sparse matrix.
 * @param outfilename [IN]    Name of the file to which to write statistics 
 *                            collected during benchmark run(s).
 * @return Pointer to freshly allocated and initialized my_HPCC_Params object.
 */
struct my_HPCC_Params*
my_HPCC_Params__construct_standalone (const int maxmem, const double fill, 
				      char* outfilename);
#else
/**
 * When we are running the SpMV benchmark in integrated mode (i.e. as part 
 * of the HPCC suite of benchmarks), this function allocates and initializes 
 * a my_HPCC_Params object, and returns a pointer to it.
 *
 * @param params [IN/OUT]     Pointer to valid HPCC_Params object.  Ownership
 *                            of the object is external; this function is not
 *                            responsible for allocation or deallocation.
 * @return Pointer to freshly allocated and initialized my_HPCC_Params object.
 */
struct my_HPCC_Params*
my_HPCC_Params__construct_integrated (HPCC_Params* params);
#endif /* SPMV_STANDALONE */


/**
 * "Destructor" of the my_HPCC_Params object.
 */
void
my_HPCC_Params__destroy (struct my_HPCC_Params* p);


/**
 * Set the failure flag in the HPCC parameters object.  The failure flag
 * indicates that the SpMV benchmark failed somehow.
 */
void
my_HPCC_Params__report_failure (struct my_HPCC_Params* p, const int failure);

/**
 * Set the maximum memory (in bytes) that the benchmark can use.
 */
void 
my_HPCC_Params__set_HPLMaxProcMem (struct my_HPCC_Params* p, 
				   const int HPLMaxProcMem);

/**
 * Returns the maximum memory (in bytes) that the benchmark can use.
 */
int 
my_HPCC_Params__get_HPLMaxProcMem (struct my_HPCC_Params* p);

/**
 * Returns the value of the fill parameter, the (approximate) fill fraction
 * of the sparse matrices to run.
 */
double 
my_HPCC_Params__get_fill (struct my_HPCC_Params* p);


/**
 * Gives the HPCC parameters object a deep copy of the benchmark results
 * object pointed to by "results".
 */
void
my_HPCC_Params__copy_benchmark_results (struct my_HPCC_Params* p, 
					struct SMVM_benchmark_results* results);


/**
 * Gives the HPCC parameters object a shallow copy of the benchmark results
 * object pointed to by "results".  Ownership of the results object passes to 
 * the HPCC parameters object.
 */
void
my_HPCC_Params__set_benchmark_results (struct my_HPCC_Params* p,
				       struct SMVM_benchmark_results* results);

/**
 * Returns the name of the output file.  Do not modify or deallocate
 * this string!
 */
const char* const 
my_HPCC_Params__get_output_filename (const struct my_HPCC_Params* p);

/**
 * Sets the output filename (shallow copy).
 */
void
my_HPCC_Params__set_output_filename (struct my_HPCC_Params* p, char* outfilename);



#endif /* _hpcc_h */
