/**
 * @file smvm_hpcc_interface.c
 * @author mfh
 * @date 2004 Dec 01
 */
#include "config.h"
#include "smvm_hpcc_interface.h"
#include "hpcc.h"
#include "smvm_benchmark_results.h"
#include "smvm_invoker.h"
#include <smvm_malloc.h>
#include <smvm_util.h>

#include <assert.h>
#include <math.h>

#ifndef SPMV_STANDALONE
#  include <mpi.h>
#endif /* SPMV_STANDALONE */


/* Element of an array that stores both the value and the original position 
 * of that value in the array.  This allows us to sort the array on the 
 * values, without forgetting the original positions of the values. */
struct int_double_pair
{
  int i;  
  double d;
};


/**
 * Comparison function for qsort; compares two (int,double) pairs in 
 * increasing order by their double values.
 */
static int
compare_int_double_pair_by_double (const void *p1, const void *p2)
{

  const double a = ((struct int_double_pair*) p1)->d;
  const double b = ((struct int_double_pair*) p2)->d;

  if (a < b)
    return -1;
  else if (a > b)
    return +1;
  /* else */
  return 0;
}

#if 0
static int
compare_double (const void *a, const void *b)
{
  const double x = *((double*)a);
  const double y = *((double*)b);

  if (x < y)
    return -1;
  else if (x > y)
    return +1;
  /* else */
  return 0;
}
#endif

/**
 * If the given array has an odd number of elements, returns (k,val), where
 * k is the index of the median in the array, and val is the value of the 
 * median.  If the given array has an even number of elements, this function 
 * does something non-standard: it takes the worst of the two middle elements,
 * and returns its index and value.  If the given array has zero elements, the
 * return value is undefined.
 *
 * @param data [IN/OUT]  Array of data.  Will be sorted on exit.
 * @param numdata [IN]   Number of elements in the data[] array.
 *
 * @return (k,val) pair
 */
static struct int_double_pair
find_median_and_index (double data[], const int numdata)
{
  struct int_double_pair p;
  struct int_double_pair *pairdata;
  int i;

  p.i = 0;
  p.d = 0.0;

  if (numdata <= 0) return p;  /* sanity check */

  /* So that we can keep track of the positions of elements of the array, make
   * a copy of data[] that stores the indices as well.  Then we can sort by the
   * doubles in the pairs. */
  pairdata = smvm_malloc (numdata * sizeof (struct int_double_pair));
  for (i = 0; i < numdata; i++)
    {
      pairdata[i].i = i;
      pairdata[i].d = data[i];
    }

  qsort (pairdata, numdata, sizeof (struct int_double_pair), 
	 compare_int_double_pair_by_double);
  /* 
   * The odd case is the easiest: get the middle element.  In the even case, 
   * we take the worse of the two middle elements, which is the lower-index
   * element in the sorted version of the array.  So in both cases, we take
   * the element with the index numdata/2 in the sorted array. 
   */
  p = pairdata[numdata/2]; /* return a copy */

  smvm_free (pairdata);
  return p;
}

/**
 * Returns (k,val), where val is the maximum element in the data[] array, and 
 * k is its index.  If the array is empty, return value is undefined.
 *
 * @param data [IN]  Array of data
 * @param numdata [IN]  Number of elements in data[]
 *
 * @return (k,val), where val is the maximum element in the data[] array, and 
 * k is its index.  If the array is empty, return value is undefined.
 */
struct int_double_pair
find_max_and_index (double data[], const int numdata)
{
  struct int_double_pair p;
  double max_val = -HUGE_VAL;
  int max_idx = 0;
  int i;

  for (i = 0; i < numdata; i++)
    {
      if (data[i] > max_val)
	{
	  max_val = data[i];
	  max_idx = i;
	}
    }
  p.i = max_idx;
  p.d = max_val;
  return p;
}



/**
 * Given an array data[0:numdata-1], returns the index k such that data[0:k]
 * are significantly higher than data[k+1:numdata-1].  It does this by finding
 * the index k at which data[k+1]-data[k] is the least. 
 *
 * @param index [OUT]   Index k that partitions data array 
 * @param data [IN]     Array of data (performance in Mflop/s)
 * @param numdata [IN]  Number of elements in the data[] array
 *
 * @return Nonzero if error; else zero.  In case of error, index is unchanged.
 */
static int
partition_data_set (int *index, double data[], const int numdata)
{
  int i;
  double min_diff = HUGE_VAL;
  double min_diff_index = -1;

  if (numdata <= 0)
    return -1; /* Empty array */
  else if (numdata == 1)
    {
      *index = 0;
      return 0;  /* special case of 1-element array */
    }

  for (i = 0; i < numdata-1; i++)
    {
      double diff = data[i+1] - data[i];

      fprintf (stderr, "i=%d, diff=%g\n", i, diff);

      if (diff < min_diff)
	{
	  min_diff = diff;
	  min_diff_index = i;
	}
    }

  if (min_diff == HUGE_VAL) return -2;

  *index = min_diff_index;
  return 0;
}



void
smvm_output_final_benchmark_results (
	struct SMVM_benchmark_results_array* ra, 
	FILE* output)
{
  struct SMVM_benchmark_results *results = NULL;
  struct int_double_pair pair1, pair2;
  double *values;
  int partition = 0;
  int i, n, errcode;

  n      = smvm_benchmark_results_array_length (ra);
  values = smvm_malloc (n * sizeof (double));

  /*
   * Make an array of the "best performances" corresponding to each problem 
   * size tested.
   */
  for (i = 0; i < n; i++)
    {
      results = get_from_smvm_benchmark_results_array (ra, i);
      values[i] = results->best_mflops_rate;
    }

  /* 
   * Partition the array of "best performances," by finding the place where 
   * the performance drops the most.  For partition==k, values[0:k] and 
   * values[k+1:n-1] is the partition.
   */
  errcode = partition_data_set (&partition, values, n);

  if (errcode != 0)
    {
      fprintf (stderr, "*** smvm_output_final_benchmark_results: Failed to "
	       "partition array of best performances! error code %d ***\n",
	       errcode);
      die (-1);
    }

  /*
   * For the small problem size group, take the best (highest) performance as 
   * the "small matrix" performance value.  For the medium problem size group, 
   * take the median performance as the "medium matrix" performance value.
   */
  pair1 = find_max_and_index (values, n);
  pair2 = find_median_and_index (&values[partition+1], n - (partition + 1));
  pair2.i += partition + 1;  /* Correct for shifting the indices up */


  printf ("\n");
  printf ("=============================================================\n");
  printf ("Sparse matrix-vector multiplication (SpMV) benchmark results:\n");
  printf ("-------------------------------------------------------------\n");
  printf ("Matrix type        Matrix size        Mflop/s        (r,c)   \n");
  printf ("-------------------------------------------------------------\n");

  results = get_from_smvm_benchmark_results_array (ra, pair1.i);
  printf ("Best case          small              %g         (%d,%d) \n", 
	  results->best_mflops_rate, results->best_r, results->best_c);
  printf ("FEM (blocked)      small              %g                     \n", 
	  results->aggregate_fem_mflops_rate);
  printf ("Non-blocked        small              %g           (1,1)   \n", 
	  results->one_by_one_mflops_rate);
  printf ("(small matrix is %d x %d with %g fill)\n",
	  results->n, results->n, results->best_fill);
  printf ("-------------------------------------------------------------\n");

  results = get_from_smvm_benchmark_results_array (ra, pair2.i);
  printf ("Best case          medium             %g         (%d,%d) \n", 
	  results->best_mflops_rate, results->best_r, results->best_c);
  printf ("FEM (blocked)      medium             %g                     \n", 
	  results->aggregate_fem_mflops_rate);
  printf ("Non-blocked        medium             %g           (1,1)     \n", 
	  results->one_by_one_mflops_rate);
  printf ("(medium matrix is %d x %d with %g fill)\n",
	  results->n, results->n, results->best_fill);
  printf ("=============================================================\n");


  smvm_free (values);
}



/** 
 * Set the appropriate error flag in the HPCC_Params struct.
 */
void
smvm_hpcc_report_failure (struct my_HPCC_Params* params, const int failure)
{
  my_HPCC_Params__report_failure (params, failure);
}


#ifndef SPMV_STANDALONE
/**
 * Create and return an MPI datatype for struct SMVM_benchmark_results.  Only
 * the address displacements of the fields in r are used; the object itself is
 * not modified, though r must point to a valid object.
 */
void
smvm_build_benchmark_results_mpi_datatype(struct SMVM_benchmark_results* r,
					  MPI_Datatype* p_type)
{
  const int num_fields = 16;  /* How many fields the struct has */
  const int num_ints   = 9;   /* Struct has all ints first, then doubles. */
  int block_lengths[16];      /* Num of elems in each "block" of new type */
  MPI_Aint displacements[16]; /* displ. of ea. elem from start of new type */
  MPI_Datatype typelist[16];  /* List of MPI types of elements */
  MPI_Aint start_address, address;
  int i;

  for (i = 0; i < num_fields; i++)
    block_lengths[i] = 1;

  for (i = 0; i < num_ints; i++)
    typelist[i] = MPI_INT;
  for (i = num_ints; i < num_fields; i++)
    typelist[i] = MPI_DOUBLE;

  /* Calculate the displacements of the fields in the struct */

  MPI_Address (&(r->failure), &start_address);
  displacements[0] = 0;

  MPI_Address (&(r->errcode), &address);
  displacements[1] = address - start_address;

  MPI_Address (&(r->n), &address);
  displacements[2] = address - start_address;

  MPI_Address (&(r->best_r), &address);
  displacements[3] = address - start_address;

  MPI_Address (&(r->best_c), &address);
  displacements[4] = address - start_address;

  MPI_Address (&(r->best_block_size), &address);
  displacements[5] = address - start_address;

  MPI_Address (&(r->worst_r), &address);
  displacements[6] = address - start_address;

  MPI_Address (&(r->worst_c), &address);
  displacements[7] = address - start_address;

  MPI_Address (&(r->worst_block_size), &address);
  displacements[8] = address - start_address;

  MPI_Address (&(r->best_mflops_rate), &address);
  displacements[9] = address - start_address;

  MPI_Address (&(r->best_fill), &address);
  displacements[10] = address - start_address;

  MPI_Address (&(r->worst_mflops_rate), &address);
  displacements[11] = address - start_address;

  MPI_Address (&(r->worst_fill), &address);
  displacements[12] = address - start_address;

  MPI_Address (&(r->one_by_one_mflops_rate), &address);
  displacements[13] = address - start_address;

  MPI_Address (&(r->one_by_one_fill), &address);
  displacements[14] = address - start_address;

  MPI_Address (&(r->aggregate_fem_mflops_rate), &address);
  displacements[15] = address - start_address;

  /* Create the MPI datatype */
  MPI_Type_struct (num_fields, block_lengths, displacements, typelist, p_type);

  /* Commit the newly created type */
  MPI_Type_commit (p_type);
}
#endif /* NOT SPMV_STANDALONE */

int
smvm_hpcc_interface (struct my_HPCC_Params* params)
{
  /* Interface modeled from the STREAM benchmark interface to HPCC.  Since our
   * benchmark is so memory-intensive, it should only be run on one processor;
   * otherwise results could be affected by memory system contention on an SMP
   * node or by a coherency protocol for a distributed shared memory machine.
   * This corresponds to `SingleStream' for the STREAM benchmark in HPCC.  */

  /* 
   * num_trials:  Number of times to run each inner loop of the SpMV benchmark
   * rand_scaling:  For transforming the output of rand() to a double
   * mem_in_kb_for_benchmark:  Amount of memory in KB to use for the current
   *                           iteration of the benchmark.  Start with 128K and
   *                           go up from there.
   */
  const int num_trials = 3; 
  double rand_scaling = 1.0 / RAND_MAX; 
  int mem_in_kb_for_benchmark = 64;

#ifndef SPMV_STANDALONE
  int my_rank, benchmark_rank, num_procs;
  MPI_Comm comm = MPI_COMM_WORLD;
  MPI_Datatype smvm_benchmark_results_t;
#endif /* NOT SPMV_STANDALONE */
  FILE* output = stdout; /* terse output file */

  struct SMVM_invoker_traits          traits;
  struct SMVM_memory_hierarchy_traits memtraits;
  struct SMVM_benchmark_results*      results = NULL;
  struct SMVM_benchmark_results_array results_array;
  
  /*-----------------------------------------------------------------------*/
  /* Code start                                                            */
  /*-----------------------------------------------------------------------*/

  WITH_DEBUG( fprintf(stderr, "=== smvm_hpcc_interface ===\n") );

  init_smvm_benchmark_results_array (&results_array);

#ifndef SPMV_STANDALONE
  MPI_Comm_size (comm, &num_procs);
  MPI_Comm_rank (comm, &my_rank);

  /* Build datatype for communicating benchmark results */
  {
    struct SMVM_benchmark_results r;
    smvm_build_benchmark_results_mpi_datatype (r, &smvm_benchmark_results_t);
  }

  /* 
   * Initialize the random number generator, to figure out which processor 
   * will run the benchmark.  I'm anxious about the fact that srand() is 
   * generally not thread-safe, but proc 0 is the only one calling the 
   * function and selecting benchmark_rank, so there's hopefully nothing
   * to worry about (but what if the individual procs are dual-core???). 
   */
  if (my_rank == 0)
    {
      srand (time (NULL));

      /* Select a single node at random (but not node 0) on which to run the 
       * benchmark. */
      do { benchmark_rank = num_procs * rand_scaling * rand(); } 
      while (benchmark_rank == 0 || benchmark_rank >= num_procs);
    }

  /* Broadcast the rank selected on node 0 to all procs */
  MPI_Bcast (&benchmark_rank, 1, MPI_INT, 0, comm); 
#endif /* NOT SPMV_STANDALONE */

#ifdef SPMV_STANDALONE
  while (mem_in_kb_for_benchmark <= my_HPCC_Params__get_HPLMaxProcMem (params) / 1024)
    {
      /* Allocate space for benchmark results */
      results = smvm_alloc_benchmark_results ();

      smvm_init_memory_hierarchy_traits (&memtraits, mem_in_kb_for_benchmark);
      smvm_construct_invocation (&traits, num_trials, 
	my_HPCC_Params__get_fill (params), memtraits, output, 
	(char*) my_HPCC_Params__get_output_filename (params), 0, 0);
      smvm_invoke_with_results (&traits, results);

      /* Check if the benchmark failed */
      if (results->failure) smvm_hpcc_report_failure (params, results->failure);

      /* Pass back results via the HPCC params object */
      my_HPCC_Params__set_benchmark_results (params, results);

      /* Output benchmark results */
      /*smvm_output_benchmark_results (results, output);*/

      append_to_smvm_benchmark_results_array (&results_array, results);

      mem_in_kb_for_benchmark *= 2;
    }
  /* Just as an experiment, let's try some "large memory model" test cases, 
   * where the source vector length is large but the fill is low.  In fact,
   * for a given source vector length n and memory usage upper bound T, the
   * required fill is:
   *
   *     T/n^2 - (2D+I)/n
   * f = ----------------
   *          D + I
   *
   * where D = sizeof(double) and I = sizeof(int).
   */
#if 0
  {
    const int I = sizeof(int);
    const int D = sizeof(double);
    int T;
    double fill = 0.001;
    int n = 1 << 19;
    
    /* Allocate as much memory as we can */
    T = my_HPCC_Params__get_HPLMaxProcMem (params);
    fill = ((double) T / (double) n - (double) (2*D+I)) / ((double) n * (double) (D+I));
    fprintf (stderr, "fill = %g\n", fill);
    fprintf (stderr, "nnz = %g\n", fill * n * n);
    assert (fill >= 0.0 && fill <= 1.0);

    smvm_init_memory_hierarchy_traits (&memtraits, T);
    smvm_construct_invocation (&traits, num_trials, fill, memtraits, output, 
    	(char*) my_HPCC_Params__get_output_filename (params), 0, 0);
    fprintf (stderr, "Actual NNZ = %d\n", traits.NNZ);
    /* smvm_invoke_with_results (&traits, results); */

    /* Check if the benchmark failed */
    /* if (results->failure) smvm_hpcc_report_failure (params, results->failure); */

    /* Output benchmark results */
    /* smvm_output_benchmark_results (results, output); */
  }
#endif

  /* TODO: Examine results for the various memory sizes.  Differentiate "small
   * memory" results from "medium memory" results, by looking at the
   * distribution of the best results for each memory size. */
  smvm_output_final_benchmark_results (&results_array, output);

  /* Close the output file */
  if (output != stdout && output != stderr)
    if (fclose (output) != 0)
      {
	fprintf (stderr, "*** ERROR:  Failed to close output file %s ***\n", 
		 my_HPCC_Params__get_output_filename (params));
	return -1;
      }

#else
  /* If I'm the proc to do the benchmark, then do it. */
  while (mem_in_kb_for_benchmark <= my_HPCC_Params__get_HPLMaxProcMem (params))
    {
      if (my_rank == benchmark_rank)
	{
	  smvm_init_memory_hierarchy_traits (&memtraits, 
	    mem_in_kb_for_benchmark);
	  smvm_construct_invocation (&traits, num_trials, 
	    my_HPCC_Params__get_fill (params), memtraits, output, 
	    (char*) my_HPCC_Params__get_output_filename (params), 0, 0);
	  smvm_invoke_with_results (&traits, results);

	  mem_in_kb_for_benchmark *= 2;
	}

      /* Broadcast benchmark results to all procs */
      MPI_Bcast (results, 1, smvm_benchmark_results_t, benchmark_rank, comm);
    }

  /* Check if the benchmark failed */
  if (results->failure) smvm_hpcc_report_failure (params, results->failure);

  /* Pass back results via the HPCC params object */
  my_HPCC_Params__set_benchmark_results (params, results);

  /* Output results to file */
  BEGIN_IO( my_rank, my_HPCC_Params__get_output_filename (params), output );
  if (my_rank == 0)
    {
      fprintf (output, "BSR SpMV benchmark, performed on node %d:\n", 
	       benchmark_rank);
      smvm_output_benchmark_results (results, output);

      if (output != stdout && output != stderr)
	if (fclose (output) != 0)
	  {
	    fprintf (stderr, "*** ERROR:  Failed to close output file %s ***\n", 
		     my_HPCC_Params__get_output_filename (params));
	    return -1;
	  }
    }
  END_IO( my_rank, output );
#endif /* SPMV_STANDALONE */

  return results->errcode;
}




