/**
 * @file smvm_hpcc_interface_tester.c
 * @author mfh
 * @date 2004 Dec 23
 *
 * Driver for testing the high-level interface to the SpMV benchmark.  The
 * interface tested is that exposed by smvm_hpcc_interface.h, which is the
 * same interface which would be used by the HPC Challenge benchmark suite
 * when it interacts with the SpMV benchmark.  This driver is only to be 
 * used when the SPMV_STANDALONE build option is enabled.
 */
#include "config.h"
#include <get_options.h>
#include "hpcc.h"
#include "smvm_hpcc_interface.h"


/*==========================================================================*/
/*                                  MAIN                                    */
/*==========================================================================*/
int
main (int argc, char *argv[])
{
  extern void smvm_set_debug_level_from_environment ();

  struct my_HPCC_Params *params;
  struct arginfo *arglist = NULL; 
  struct arginfo *arg;
  char* outfilename;
  int maxmem_in_KB;
  int default_max_mem = 1024;
  /* default_fill:  median fill from Matrices 2-9 in Rich Vuduc's thesis */
  double default_fill = 0.0022287;
  double fill;

  smvm_set_debug_level_from_environment ();

  arglist = register_arginfo (arglist, 'f', STRINGARG, "", "filename in "
			      "which to store statistics of the benchmark run");
  arglist = register_arginfo (arglist, 'm', INTARG, &default_max_mem, 
			      "max amount of memory (in KB) to allow for use "
			      "by the benchmark");
  arglist = register_arginfo (arglist, 'l', DOUBLEARG, &default_fill, 
			      "fill fraction (between 0.0 and 1.0 inclusive)");
  get_options (argc, argv, arglist, NULL);



  /* Look for the filename argument.  If we don't have it, dump usage and 
   * abort.  Otherwise get the filename. */
  if (!got_arg_value_p (arglist, 'f'))
    {
      dump_usage (stderr, argv[0], arglist, NULL);
      return -1;
    }
  outfilename = get_string_argval (find_arginfo (arglist, 'f'));
  fill = get_double_argval (find_arginfo (arglist, 'l'));

  /* Look for the max-mem argument.  If we don't have it, give a reasonable
   * default.  Do some range-checking here too.  */
  if (got_arg_p (arglist, 'm') && !got_arg_value_p (arglist, 'm'))
    maxmem_in_KB = 1024;
  else
    {
      arg = find_arginfo (arglist, 'm');
      maxmem_in_KB = get_int_argval (arg);
      /* Range-check */
      if (maxmem_in_KB <= 0 || maxmem_in_KB > 8000000)
	{
	  fprintf (stderr, "*** ERROR: -m parameter %d is out of range "
		   "***\n\n", maxmem_in_KB);
	  dump_usage (stderr, argv[0], arglist, NULL);
	  return -1;
	}
    }

  /* Range-check the fill parameter.  If negative, use the default. */
  if (fill < 0.0)
    fill = default_fill;
  else if (fill > 1.0)
    {
      fprintf (stderr, "*** ERROR: -f parameter %g is out of range ***\n\n", fill);
      dump_usage (stderr, argv[0], arglist, NULL);
      return -1;
    }

#ifdef SPMV_STANDALONE
  params = my_HPCC_Params__construct_standalone (maxmem_in_KB * 1024, fill, outfilename);
#else
#  error "This file is only for use with SPMV_STANDALONE!"
#endif /* SPMV_STANDALONE */
  smvm_hpcc_interface (params);
  my_HPCC_Params__destroy (params);

  destroy_arginfo_list (arglist);
  return 0;
}
