/**
 * @file smvm_invoker.c
 * @author mfh
 * @date 2004 Nov 30
 ****************************************************************************/
#include "config.h"
#include "avltree.h"
#include "block_size_parser.h"
#include "smvm_memory_hierarchy_traits.h"
#include "smvm_timing_results.h"
#include "smvm_invoker.h"
#include "smvm_benchmark.h"  
#include <smvm_malloc.h>

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>   
#include <string.h>
#include "smvm_util.h"

/**
 * @brief Choose matrix size n for sparse benchmark.
 *
 * For running the SMVM benchmark with a sparse matrix.  Chooses n (the source
 * and destination vector length, and also the dimension of the matrix: n x n)
 * so that the resulting sparse matrix and source and destination vectors fit
 * within the given memory bound.  NNZ is simply fill * n, which the client is
 * responsible for calculating.
 *
 * @param max_allocable_mem [IN]    Maximum allocable memory (in B) available 
 *                                  for the sparse matrix data structure and 
 *                                  the source and destination vectors.
 * @param fill [IN]                 Desired fill of the matrix.
 * @param p_errcode [OUT]           Reports any errors that may have occurred
 *                                  in this function.  
 * @return n
 */
int 
smvm_pick_sparse_data_size (const int max_allocable_mem, const double fill, 
			    enum SMVM_problem_code* p_errcode);


/**
 * @brief Choose matrix size dense_n for dense-in-sparse-format benchmark.
 *
 * For running the SMVM benchmark with a dense matrix in sparse format.
 * Chooses n (the source and destination vector length, and also the dimension
 * of the matrix: n x n) so that the resulting matrix and source and
 * destination vectors fit within the given memory bound.  
 *
 * @param max_allocable_mem [IN]    Maximum allocable memory (in B) available 
 *                                  for the sparse matrix data structure and 
 *                                  the source and destination vectors.
 * @param p_errcode [OUT]           Reports any errors that may have occurred
 *                                  in this function.  
 * @return n
 */
int
smvm_pick_dense_data_size (const int max_allocable_mem, 
			   enum SMVM_problem_code* p_errcode);
 


/*****************************************************************************/
double 
smvm_fill (int NNZ, int m, int n)
{
  /* If m and n are particularly large, computing their product as an int could
   * overflow.  (I've seen this happen before, for real-life examples.)  So we
   * convert them both to doubles first.
   */
  double d_m = (double) m;
  double d_n = (double) n;
  double d_NNZ = (double) NNZ;

  return d_NNZ / (d_m * d_n);
}


/*****************************************************************************/
int
smvm_bytes_used_by_BSR_matrix (int NNZ, int m, int n, int r, int c)
{
  int D = sizeof (double);
  int I = sizeof (int);

  return (D * NNZ) + (I * NNZ / (r*c)) + (I * m / r);
}


/*****************************************************************************/
int 
smvm_pick_sparse_data_size (const int max_allocable_mem, const double fill, 
			    enum SMVM_problem_code* p_errcode)
{
  /* 
   * MFH 2004 Mar 17: Since we no longer need to pick source and destination
   * vector length (i.e. n) in order to satisfy in-cache constraints, we have
   * only one constraint:  the upper bound U on total memory expenditure.
   * Total memory use is:
   *
   * @fn[
   *    NNZ (I + D) + n(I + 2D),
   * @fn]
   *
   * where I = sizeof(int) and D = sizeof(double).  There is only one
   * constraint, but two variables to fix (n and NNZ).  So, we introduce a
   * fill constant f: NNZ = f * n * n, chosen to represent common matrices
   * encountered in practice.  Setting the above formula equal to U, a 
   * quadratic for n must then be solved.  We take the positive branch of
   * the quadratic formula.
   *
   * U is measured in KB and is likely O(10^6) or more, so we need to be
   * careful of integer overflow when computing n (which is measured in single
   * elements).  In particular, we carry out calculations in floating-point
   * arithmetic and then round down.
   */
  enum SMVM_problem_code errcode = NO_PROBLEMS;
  double D     = (double) sizeof (double);
  double I     = (double) sizeof (int);
  double U;     
  int    n;     

  if (max_allocable_mem <= 0)
    {
      fprintf (stderr, "*** ERROR *** \n"
	       "smvm_pick_sparse_data_size:\n"
	       "max_allocable_mem = %d KB, which is non-positive!\n", 
	       max_allocable_mem);
      die (-1);
    }
  if (fill < 0.0 || fill > 1.0)
    {
      fprintf (stderr, "*** ERROR *** \n"
	       "smvm_pick_sparse_data_size:\n"
	       "fill = %g, which is not in [0,1]!\n", fill);
      die (-1);
    }

  WITH_DEBUG( fprintf(stderr, "smvm_pick_sparse_data_size:  "
		      "max_allocable_mem = %d KB, fill = %g\n", 
		      max_allocable_mem, fill) );

  /* max_allocable_mem measured in KB, but U is number of bytes. */
  U = ((double) 1024) * ((double) max_allocable_mem);
  WITH_DEBUG( fprintf(stderr, "\tU = %g\n", U) );

  /* Solve the quadratic equation for n */
  {
    double a = fill * (I + D);
    double b = I + 2.0*D;
    double c = -U;

    /* n as a double, rather than an int. */
    /* FIXME: use a more accurate quadratic formula! */
    double double_n = (-b + sqrt (b*b - 4*a*c)) / (2.0*a);

    if (double_n <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_pick_sparse_data_size:\n"
		 "resulting n = %.0g is non-positive!\n", double_n);
	die (-1);
      }

    /* Check for overflow before committing to matrix dimension n. */
    if (double_n > (double) INT_MAX)
      {
	WITH_WARN( fprintf(stderr,
			   "While picking sparse data set size: optimally "
			   "large n = %.0g too large for int; choosing "
			   "smaller n to fit...\n", double_n) );
	while ((double_n > (double) INT_MAX) && (double_n > 0.0))
	  double_n = double_n / 2.0;
	
	WITH_WARN( fprintf(stderr, "New n:  %d\n", (int) double_n) );
	WITH_DEBUG( fprintf(stderr, "Sparse data size selection:  "
			    "Fixed n to %d to fit in int\n", (int) double_n) );
      }
  
    n = (int) double_n;
  }

  *p_errcode = errcode;
  return n;
}


/*****************************************************************************/
int
smvm_pick_dense_data_size (const int max_allocable_mem, 
			   enum SMVM_problem_code* p_errcode)
{
  /* A dense matrix in sparse format is a sparse matrix with fill 1. 
   * So we use the same formulae as above, but set fill = 1. 
   */
  enum SMVM_problem_code errcode = NO_PROBLEMS;
  double D     = (double) sizeof (double);
  double I     = (double) sizeof (int);
  double U;     
  int    n;     

  if (max_allocable_mem <= 0)
    {
      fprintf (stderr, "*** ERROR *** \n"
	       "smvm_pick_dense_data_size:\n"
	       "max_allocable_mem = %d KB, which is non-positive!\n", 
	       max_allocable_mem);
      die (-1);
    }

  WITH_DEBUG( fprintf(stderr, "smvm_pick_dense_data_size:  "
		      "max_allocable_mem = %d KB\n", max_allocable_mem) );

  /* max_allocable_mem measured in KB, but U is number of bytes. */
  U = ((double) 1024) * ((double) max_allocable_mem);
  WITH_DEBUG( fprintf(stderr, "\tU = %g\n", U) );

  /* Solve the quadratic equation for n */
  {
    double a = (I + D);
    double b = I + 2.0*D;
    double c = -U;

    /* n as a double, rather than an int. */
    double double_n = (-b + sqrt (b*b - 4*a*c)) / (2.0*a);

    if (double_n <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_pick_dense_data_size:\n"
		 "resulting n = %.0g is non-positive!\n", double_n);
	die (-1);
      }

    /* Check for overflow before committing to matrix dimension n. */
    if (double_n > (double) INT_MAX)
      {
	WITH_WARN( fprintf(stderr,
			   "While picking dense data set size: optimally "
			   "large n = %.0g too large for int; choosing "
			   "smaller n to fit...\n", double_n) );
	while ((double_n > (double) INT_MAX) && (double_n > 0.0))
	  double_n = double_n / 2.0;
	
	WITH_WARN( fprintf(stderr, "New n:  %d\n", (int) double_n) );
	WITH_DEBUG( fprintf(stderr, "Dense data size selection: "
			    "Fixed n to %d to fit in int\n", 
			    (int) double_n) );
      }
  
    n = (int) double_n;
  }

  *p_errcode = errcode;
  return n;
}


/*****************************************************************************/
void
smvm_dense_single_invocation (struct SMVM_invoker_traits* p_traits, int r, 
			      int c, struct SMVM_timing_results* p_results)
{
  struct SMVM_parameters smvm_params;

  /* ----------------------------------------------------------------- */
  /* ------------ Sanity checks and setting parameters --------------- */
  /* ----------------------------------------------------------------- */

    
  if (p_traits->dense_n <= 0)
    {
      die_with_message ("*** ERROR:  smvm_dense_single_invocation:  "
			"dense_n out of range! ***", -1);
    }
  else 
    {
      smvm_params.m = p_traits->dense_n;
      smvm_params.n = p_traits->dense_n;
    }
  if (r <= 0)
    {
      die_with_message ("*** ERROR:  smvm_dense_single_invocation:  r "
			"out of range! ***", -1);
    }
  else
    smvm_params.r = r;

  if (c <= 0)
    {
      die_with_message ("*** ERROR:  smvm_dense_single_invocation:  c "
			"out of range! ***", -1);
    }
  else
    smvm_params.c = c;

  if (p_traits->num_trials < 0)
    {
      die_with_message ("*** ERROR:  smvm_dense_single_invocation:  "
			"num_trials out of range! ***", -1);
    }
  else
    smvm_params.num_trials = p_traits->num_trials;

  /* The benchmark wants a percentage rather than a fraction. */
  smvm_params.percent_fill = 100.0;

  /* Results go only to stdout. */
  smvm_params.dataoutfile = stdout;

  smvm_benchmark_with_results (&smvm_params, p_results);

  /*
   * @bug This (parent/caller, not `smvm_benchmark()') function should return
   * an error code corresponding to any errors that the function below may
   * return.
   */
}


/*****************************************************************************/
void 
smvm_sparse_single_invocation (struct SMVM_invoker_traits* p_traits, int r, 
			       int c, struct SMVM_timing_results* p_results)
{
  struct SMVM_parameters smvm_params;

  WITH_DEBUG( fprintf(stderr, "smvm_sparse_single_invocation:\n") );

  /* ----------------------------------------------------------------- */
  /* ------------ Sanity checks and setting parameters --------------- */
  /* ----------------------------------------------------------------- */

    
  if (p_traits->n <= 0)
    {
      die_with_message ("*** ERROR:  smvm_sparse_single_invocation:  n out of "
			"range! ***", -1);
    }
  else 
    {
      smvm_params.m = p_traits->n;
      smvm_params.n = p_traits->n;
    }
  if (r <= 0)
    {
      die_with_message ("*** ERROR:  smvm_sparse_single_invocation:  r "
			"out of range! ***", -1);
    }
  else
    smvm_params.r = r;

  if (c <= 0)
    {
      die_with_message ("*** ERROR:  smvm_sparse_single_invocation:  c "
			"out of range! ***", -1);
    }
  else
    smvm_params.c = c;

  if (p_traits->num_trials < 0)
    {
      die_with_message ("*** ERROR:  smvm_sparse_single_invocation:  "
			"num_trials out of range! ***", -1);
    }
  else
    smvm_params.num_trials = p_traits->num_trials;

  if (p_traits->fill < 0.0 || p_traits->fill > 1.0)
    {
      die_with_message ("*** ERROR:  smvm_sparse_single_invocation:  "
			"fill out of range! ***\n", -1);
    }
  else
    {
      /* The benchmark wants a percentage rather than a fraction. */
      smvm_params.percent_fill = 100.0 * p_traits->fill;
    }

  /* Results go only to stdout. */
  smvm_params.dataoutfile = stdout;

  smvm_benchmark_with_results (&smvm_params, p_results);

  /*
   * @bug Should return an error code corresponding to any errors that the
   * function below may return.
   */
}



/*****************************************************************************/
void
smvm_construct_invocation(struct SMVM_invoker_traits* p_traits, 
			  int num_trials, 
			  double fill,
			  struct SMVM_memory_hierarchy_traits memtraits, 
			  FILE* output, char *raw_output_filename,
			  int outputlevel, int debuglevel)
{
  enum SMVM_problem_code sparse_errs, dense_errs;
  int dbg = debuglevel;

  WITH_DEBUG( fprintf (stderr, "smvm_construct_invocation:\n") );

  /* Sanity checks. */
  if (p_traits == NULL)
    die_with_message ("*** ERROR: smvm_construct_invocation: "
		      "SMVM_invoker_traits object is NULL! ***\n", -1);
  else if (num_trials < 0)
    die_with_message ("*** ERROR: smvm_construct_invocation: Number of "
		      "trials cannot be negative! ***\n", -1);
  else if (fill < 0.0)
    fill = 0.01; /* default */
  else if (fill > 1.0)
    die_with_message ("*** ERROR: smvm_construct_invocation: invalid "
		      "fill fraction ***\n", -1);
  else if (output == NULL)
    die_with_message ("*** ERROR: smvm_construct_invocation: Output file "
		      "is NULL!!! ***\n", -1);

  p_traits->num_trials  = num_trials;
  p_traits->memtraits   = memtraits;
  p_traits->output      = output;
  p_traits->raw_output_filename = raw_output_filename;
  p_traits->outputlevel = outputlevel;
  p_traits->debuglevel  = debuglevel;

  p_traits->fill = fill;
  p_traits->n = smvm_pick_sparse_data_size (memtraits.max_allocable_mem, fill, 
					    &sparse_errs);  
  p_traits->NNZ = fill * (p_traits->n) * (p_traits->n);
  p_traits->dense_n = smvm_pick_dense_data_size (memtraits.max_allocable_mem, 
						 &dense_errs);  

  /* 
   * @todo Print a more informative error message about which problems
   * occurred, or log to some file or something.
   */
  if (sparse_errs != NO_PROBLEMS)
    fprintf (stderr, "*** WARNING ***\n"
	     "\tData size selelection problems for sparse benchmark.\n"
	     "\tContinuing anyway...\n");
  if (dense_errs != NO_PROBLEMS)
    fprintf (stderr, "*** WARNING ***\n"
	     "\tData size selelection problems for dense-in-sparse-format "
	     "benchmark!\n\tContinuing anyway...\n");

  WITH_DEBUG( fprintf (stderr, "\nResults of picking sparse matrix size "
		       "parameters:\n\tn       = %d\n"
		       "\tNNZ     = %d\n"
		       "\tfill    = %f\n"
		       "\tdense_n = %d\n",
		       p_traits->n, p_traits->NNZ, p_traits->fill, 
		       p_traits->dense_n) );

  /* More sanity checks. */
  if (p_traits->n <= 0)
    {
      fprintf (stderr, 
	       "*** ERROR:  smvm_construct_invocation:  n = %d <= 0 ***\n", 
	       p_traits->n);
      die (-1);
    }
  if (p_traits->dense_n <= 0)
    {
      fprintf (stderr, 
	       "*** ERROR:  smvm_construct_invocation:  "
	       "dense_n = %d <= 0 ***\n", p_traits->dense_n);
      die (-1);
    }
  if (p_traits->NNZ <= 0)
    {
      fprintf (stderr, "*** ERROR:  smvm_construct_invocation:  "
	       "NNZ = %d <= 0 ***\n", p_traits->NNZ);
      die (-1);
    }
  /* n^2 could overflow int, so we cast to double. */
  if (((double) p_traits->NNZ) > 
      ((double)(p_traits->n)) * ((double)(p_traits->n)))
    {
      fprintf (stderr, "*** ERROR:  smvm_construct_invocation:  "
	       "NNZ = %d > n^2, which is impossible ***\n", p_traits->NNZ);
      die (-1);
    }
}

/**
 * Prints a line of column labels for verbose output.  Not to be called by 
 * users.
 *
 * @see smvm_print_verbose_results
 */
static void
smvm_print_verbose_results_header (FILE* out)
{
  fprintf (out, "m,n,r,c,nnzb,num_trials,t_median,t_min,t_max,mflops,"
	   "num_loads,num_stores\n");  
}

int
smvm_print_verbose_results (char *filename, 
			    struct SMVM_timing_results results[],
			    const int num_results)
{
  FILE *out;

  /* Don't output verbose results if filename is NULL or empty. */
  if (filename == NULL || strcmp (filename, "") == 0)
    return -1;

  /* Check if the file exists, by trying to open it for reading */
  out = fopen (filename, "r");
  if (out == NULL)
    {
      /* File doesn't exist, so try to create it */
      out = fopen (filename, "w");
      /* If we failed to create file, then return with error code */
      if (out == NULL)
	{
	  fprintf (stderr, "*** smvm_print_verbose_results: failed to open "
		   "%s for writing verbose results ***\n", filename);
	  return -1;
	}
      /* The file is open, so print column labels to it */
      smvm_print_verbose_results_header (out);
    }
  else /* close the file and open it again for appending */
    {
      if (0 != fclose (out))
	{
	  fprintf (stderr, "*** smvm_print_verbose_results: failed to close "
		   "raw output file %s ***\n", filename);
	  return -2;
	}
      out = fopen (filename, "a");
      if (out == NULL)
	{
	  fprintf (stderr, "*** smvm_print_verbose_results: failed to open "
		   "%s for writing verbose results ***\n", filename);
	  return -1;
	}
    }

  int i;
  for (i = 0; i < num_results; i++)
    {
      struct SMVM_timing_results *p = &results[i];
      fprintf (out, "%d,%d,%d,%d,%d,%d,%e,%e,%e,%e,%d,%d\n", 
	       p->m, p->n, p->r, p->c, p->nnzb, p->num_trials, p->t_median, 
	       p->t_min, p->t_max, p->mflops, p->num_loads, p->num_stores);
    }

  if (0 != fclose (out))
    {
      fprintf (stderr, "*** smvm_print_verbose_results: failed to close "
	       "raw output file %s ***\n", filename);
      return -2;
    }

  return 0;
}


/**
 * @brief Post-processes and outputs set of SMVM_timing_results.
 *
 * Part of the implementation of the benchmark invoker.  Post-processes an
 * array of SMVM_timing_results from runs of the sparse benchmark, formats the
 * results nicely, and outputs them to the given output stream.  Array of
 * results is sorted in place.
 *
 * @param results [IN,OUT]  Array of results from sparse benchmark.
 * @param num_results [IN]  Number of records in above array.
 * @param output [OUT]      Stream to which to write output.
 */
static void
smvm_process_and_output_sparse_results (struct SMVM_timing_results results[],
					const int num_results, FILE* output)
{
  /* 
   * Sort timing results (by the median) so that we can extract the best
   * and worst.  When done, the first record has the least median time.  
   * (r,c) are the identifying fields in each record.
   */
  qsort (results, num_results, sizeof(struct SMVM_timing_results),
	 compare_timing_results_by_t_median);
  /* 
   * Report results (in Mflops/s) of sparse benchmark: Show worst and best
   * median completion times, and their corresponding (r,c).  If something 
   * that looks int-like is calculated using doubles, that's because we want
   * to avoid overflow of int (which is a real possibility for large n).
   */
  {
    int i;
    int n = results[0].n;
    double n_squared = (double) n * (double) n;

    int best_r = results[0].r;
    int best_c = results[0].c;
    int best_block_size = best_r * best_c;
    int worst_r = results[num_results - 1].r;
    int worst_c = results[num_results - 1].c;
    int worst_block_size = worst_r * worst_c;

    double best_mflops_rate;
    double best_fill;
    double worst_mflops_rate;
    double worst_fill;
    double one_by_one_mflops_rate = 0.0;
    double one_by_one_fill = 0.0;

    if (n != results[num_results - 1].n)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "n = %d used for best result of sparse benchmark is "
		 "different than n = %d used\n"
		 "for worst result of sparse benchmark!\n", 
		 n, results[num_results - 1].n);
	die (-1);
      }
    if (n <= 0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "n = %d which is non-positive!\n", n);
	die (-1);
      }
    best_fill = ((double) results[0].nnzb * 
		 (double) best_block_size) / n_squared;
    worst_fill = ((double) results[num_results - 1].nnzb * 
		  (double) worst_block_size) / n_squared;

    /* Find the 1x1 case */
    for (i = 0; i < num_results; i++)
      {
	if (results[i].r == 1 && results[i].c == 1)
	  {
	    one_by_one_mflops_rate = results[i].mflops / results[i].t_median;
	    one_by_one_fill = (double) results[i].nnzb / n_squared;
	    break;
	  }
      }
    if (one_by_one_mflops_rate == 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Failed to find 1x1 mflops rate!\n");
	die (-1);
      }
    
    if (results[0].t_median <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with fastest median completion time\n"
		 "finished in non-positive median time!\n");
	die (-1);
      }
    if (results[0].mflops <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with fastest median completion time\n"
		 "involves zero measured floating point operations!\n");
	die (-1);
      }
    best_mflops_rate = results[0].mflops / results[0].t_median;

    if (results[num_results - 1].t_median <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with slowest median completion time\n"
		 "finished in non-positive median time!\n");
	die (-1);
      }
    if (results[num_results - 1].mflops <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with slowest median completion time\n"
		 "involves zero measured floating point operations!\n");
	die (-1);
      }
    worst_mflops_rate = results[num_results - 1].mflops / results[num_results - 1].t_median;

    /* 
     * Output neatly formatted results.  Note that fill differs only slightly
     * for different block dimensions, when n is large.
     */
    fprintf (output, "For %d x %d BSR sparse matrix with %g fill:\n", 
	     n, n, best_fill);
    fprintf (output, "Best median Mflop/s rate is %g, for (r,c) = (%d,%d)\n", 
	     best_mflops_rate, best_r, best_c);
    fprintf (output, "Worst median Mflop/s rate is %g, for (r,c) = (%d,%d)\n", 
	     worst_mflops_rate, worst_r, worst_c);
    fprintf (output, "1x1 median Mflop/s rate is %g\n",
	     one_by_one_mflops_rate);
  }
}


/**
 * @brief Post-processes and outputs set of SMVM_timing_results.
 *
 * Part of the implementation of the benchmark invoker.  Post-processes an
 * array of SMVM_timing_results from runs of the dense-in-sparse-format
 * benchmark, formats the results nicely, and outputs them to the given output
 * stream.  Array of results is sorted in place.
 *
 * @param results [IN,OUT]  Array of results from dense benchmark.
 * @param num_results [IN]  Number of records in above array.
 * @param output [OUT]      Stream to which to write output.
 */
void
smvm_process_and_output_dense_results (struct SMVM_timing_results results[], 
					const int num_results, FILE* output)
{
  /* 
   * Sort timing results (by the median) so that we can extract the best
   * and worst.  When done, the first record has the least median time.  
   * (r,c) are the identifying fields in each record.
   */
  qsort (results, num_results, sizeof(struct SMVM_timing_results),
	 compare_timing_results_by_t_median);
  /* 
   * Report results (in Mflops/s) of dense benchmark: Show worst and best
   * median completion times, and their corresponding (r,c).  If something 
   * that looks int-like is calculated using doubles, that's because we want
   * to avoid overflow of int (which is a real possibility for large n).
   */
  {
    int i;
    int n = results[0].n;
    double n_squared = (double) n * (double) n;

    int best_r = results[0].r;
    int best_c = results[0].c;
    int best_block_size = best_r * best_c;
    int worst_r = results[num_results - 1].r;
    int worst_c = results[num_results - 1].c;
    int worst_block_size = worst_r * worst_c;

    double best_mflops_rate;
    double worst_mflops_rate;
    double one_by_one_mflops_rate = 0.0;
    double one_by_one_fill = 0.0;

    if (n != results[num_results - 1].n)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_dense_results:\n"
		 "n = %d used for best result of dense benchmark is "
		 "different than n = %d used\n"
		 "for worst result of dense benchmark!\n", 
		 n, results[num_results - 1].n);
	die (-1);
      }
    if (n <= 0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_dense_results:\n"
		 "n = %d which is non-positive!\n", n);
	die (-1);
      }
    
    if (results[0].t_median <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_dense_results:\n"
		 "Dense result with fastest median completion time\n"
		 "finished in non-positive median time!\n");
	die (-1);
      }
    if (results[0].mflops <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_dense_results:\n"
		 "Dense result with fastest median completion time\n"
		 "involves zero measured floating point operations!\n");
	die (-1);
      }
    best_mflops_rate = results[0].mflops / results[0].t_median;

    if (results[num_results - 1].t_median <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_dense_results:\n"
		 "Dense result with slowest median completion time\n"
		 "finished in non-positive median time!\n");
	die (-1);
      }
    if (results[num_results - 1].mflops <= 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_dense_results:\n"
		 "Dense result with slowest median completion time\n"
		 "involves zero measured floating point operations!\n");
	die (-1);
      }
    worst_mflops_rate = results[num_results - 1].mflops / results[num_results - 1].t_median;

    for (i = 0; i < num_results; i++)
      {
	if (results[i].r == 1 && results[i].c == 1)
	  {
	    one_by_one_mflops_rate = results[i].mflops / results[i].t_median;
	    one_by_one_fill = (double) results[i].nnzb / n_squared;
	    break;
	  }
      }
    if (one_by_one_mflops_rate == 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Failed to find 1x1 mflops rate!\n");
	die (-1);
      }
 

    /* Output neatly formatted results. */
    fprintf (output, "For %d x %d BSR dense matrix in BSR sparse format:\n",
	     n, n);
    fprintf (output, "Best median Mflop/s rate is %g, for (r,c) = (%d,%d)\n", 
	     best_mflops_rate, best_r, best_c);
    fprintf (output, "Worst median Mflop/s rate is %g, for (r,c) = (%d,%d)\n", 
	     worst_mflops_rate, worst_r, worst_c);
    fprintf (output, "1x1 median Mflop/s rate is %g\n",
	     one_by_one_mflops_rate);
  }
}


/**
 * Struct used by copy_avl_tree_into_array_visitor.
 */
struct copy_avl_tree_into_array_t
{
  struct int_pair *array;
  int curlen;
  int maxlen;
};

/**
 * Visit function for copying an avl_tree into a sorted array.
 * Used by copy_avl_tree_into_array.  The traverse_in_order function
 * should be called in copy_avl_tree_into_array.
 */
static void
copy_avl_tree_into_array_visitor (avl_tree T, void *data)
{
  struct copy_avl_tree_into_array_t *c;

  if (c->maxlen <= c->curlen)
    {
      c->maxlen = 1 + 2*c->maxlen;
      c->array = smvm_realloc (c->array, c->maxlen * sizeof (struct int_pair));
    }
  (c->array)[c->curlen++] = T->element;
}

/**
 * Copies the contents of the given avl_tree into an array.
 *
 * @param array [OUT]   Pointer to array of pairs of ints, which were copied 
 *                      out of T.
 * @param num_elements [OUT]  Pointer to number of elements in above array.
 * @param T [IN]              Set (implemented as AVL tree) of int pairs.
 */
static void
copy_avl_tree_into_array (struct int_pair **array, int *num_elements, avl_tree T)
{
  struct copy_avl_tree_into_array_t c;
  c.array = NULL;
  c.curlen = 0;
  c.maxlen = 0;

  traverse_in_order (T, copy_avl_tree_into_array_visitor, &c);

  *array = c.array;
  *num_elements = c.curlen;
}


/**
 * For all the timing results in the given array, extracts the (r,c) pairs
 * and returns a set of the (r,c) pairs that occur.
 *
 * @param pairs [OUT]        Set (list) of (r,c) pairs
 * @param num_pairs [OUT]  Number of pairs in the above set
 * @param results [IN] 
 * @param num_results [IN]
 */
static avl_tree
extract_block_dim_pairs (const struct SMVM_timing_results results[],
			 const int num_results)
{
  avl_tree set = NULL;
  int i;

  for (i = 0; i < num_results; i++)
    {
      struct int_pair pair;

      pair.first = results[i].r;
      pair.second = results[i].c;
      set = insert (pair, set);
    }

  return set;
}


/**
 * Comparison function for block_size_data_t, for sorting by decreasing 
 * frequency.
 */
static int
compare_block_size_data (const void *a, const void *b)
{
  struct block_size_data_t *a1 = (struct block_size_data_t*) a;
  struct block_size_data_t *b1 = (struct block_size_data_t*) b;

  if (a1->frequency < b1->frequency)
    return +1; /* deliberately reversed */
  else if (a1->frequency > b1->frequency)
    return -1;
  
  return 0;
}

static void
sort_block_size_data_array_by_decreasing_frequency (struct block_size_data_t *array, 
						    const int num_elements)
{
  qsort (array, num_elements, sizeof (struct block_size_data_t),
	 &compare_block_size_data);
}

/**
 * @brief Post-process and store summary of timing results in `p_summary'.
 *
 * Part of the implementation of the benchmark invoker.  Post-processes an
 * array of SMVM_timing_results from runs of the sparse benchmark, and 
 * stores the summary in `p_summary'. Array of results is sorted in place, 
 * so not declared const.
 *
 * @param p_summary [OUT]   To be filled in with summary of benchmark results.
 * @param results [IN,OUT]  Array of results from dense benchmark.
 * @param num_results [IN]  Number of records in above array.
 * @param output [OUT]      Stream to which to write error messages.
 */
void
smvm_process_and_store_sparse_results (struct SMVM_benchmark_results* p_summary, 
				       struct SMVM_timing_results results[],
				       const int num_results, FILE* output)
{
  if (output == NULL)
    output = stderr;

  {
    avl_tree block_dims_set = NULL;
    struct block_size_data_t *data = NULL;
    int numdata = 0;
    int i;
    int can_calculate_aggregate_p = 1;
    int N = 0; /* Total of the frequencies of the top 8 block dimensions */

    p_summary->aggregate_fem_mflops_rate = 0.0;

    /* Extract the set of block dimensions encountered in the timing results */
    block_dims_set = extract_block_dim_pairs (results, num_results);

	/**
	 * \bug parse_block_sizes_file should not be called with an assert, and the
	 * name of the block sizes file should not be hardcoded.  A clear error
	 * message should be printed if the file cannot be found or if the file is
	 * empty or if data cannot be retrieved from the file.
	 */
    /* Parse the file of block sizes.  FIXME:  Should not be an assert! */
    assert (0 == parse_block_sizes_file (&data, &numdata, "block_sizes.csv"));
    assert (numdata > 0);
    assert (data != NULL);

    /* fprintf (stderr, "Got %d data from file\n", numdata); */

    /*
    for (i = 0; i < numdata; i++)
      fprintf (stderr, "Datum:  %d,%d,%d\n", data[i].r, data[i].c, data[i].frequency);
    fprintf (stderr, "\n");
    */

    sort_block_size_data_array_by_decreasing_frequency (data, numdata);
    /* Make sure that the top 8 block dimensions from the block sizes data 
     * file are represented in the benchmark results.  While we are doing
     * this, add up the frequencies, so that we can calculate weights for
     * the scores corresponding to these block dimensions. */
    for (i = 0; i < MIN(numdata, 8); i++)
      {
	struct int_pair pair;
	pair.first = data[i].r;
	pair.second = data[i].c;
	N += data[i].frequency;
	if (find (pair, block_dims_set) == NULL)
	  {
	    fprintf (stderr, "*** smvm_process_and_output_sparse_results:  "
		     "unable to find (r,c) = (%d,%d) in the benchmark "
		     "results, even though it appears in the block sizes data "
		     "file! ***\n", pair.first, pair.second);
	    can_calculate_aggregate_p = 0;
	  }
      }

    if (N <= 0 && can_calculate_aggregate_p)
      {
	fprintf (stderr, "*** smvm_process_and_output_sparse_results:  N<=0 "
		 "***\n");
	can_calculate_aggregate_p = 0;
      }

    if (can_calculate_aggregate_p)
      {
	/* Now, wouldn't it be nice if C had an associative array data structure?
	 * Unfortunately we have to go back into the benchmark results and do a 
	 * linear search for each (r,c) pair of the top 8 frequencies.  While we
	 * are at it, compute the weight for each pair. */
	for (i = 0; i < MIN(numdata, 8); i++)
	  {
	    const int r = data[i].r;
	    const int c = data[i].c;
	    int found_pair_p = 0;
	    int j;
	    double weight = (double) data[i].frequency / (double) N;
	    assert (weight >= 0.0 && weight <= 1.0);

	    /* Find the current (r,c) pair in the benchmark results */
	    /*fprintf (stderr,"\tCurrent (r,c) = (%d,%d)\n", r, c);*/
	    for (j = 0; j < num_results; j++)
	      {
		/*fprintf (stderr, "\t\t(r,c) = (%d,%d)\n", results[j].r, results[j].c);*/
		if ((results[j].r == r) && (results[j].c == c))
		  {
		    found_pair_p = 1;
		    break;
		  }
	      }
	    /* We've already guaranteed that the pair is there */
	    assert (found_pair_p); 

	    p_summary->aggregate_fem_mflops_rate += 
	      weight * (double) results[j].mflops / results[j].t_median;
	  }
      }
    else
      {
	p_summary->aggregate_fem_mflops_rate = -1.0; /* flag */
      }

    smvm_free (data);
    block_dims_set = make_empty (block_dims_set);
  }

  /* 
   * Sort timing results (by the median) so that we can extract the best
   * and worst.  When done, the first record has the least median time.  
   * (r,c) are the identifying fields in each record.
   */
  qsort (results, num_results, sizeof(struct SMVM_timing_results),
	 compare_timing_results_by_t_median);

  /* 
   * Calculate results (in Mflops/s) of sparse benchmark: worst and best
   * median completion times, and their corresponding (r,c).  If something 
   * that looks int-like is calculated using doubles, that's because we want
   * to avoid overflow of int (which is a real possibility for large n).
   */
  {
    int i;
    int n = results[0].n;
    double n_squared = (double) n * (double) n;

    int best_r = results[0].r;
    int best_c = results[0].c;
    int best_block_size = best_r * best_c;
    int worst_r = results[num_results - 1].r;
    int worst_c = results[num_results - 1].c;
    int worst_block_size = worst_r * worst_c;

    double best_mflops_rate;
    double best_fill;
    double worst_mflops_rate;
    double worst_fill;
    double one_by_one_mflops_rate = 0.0;
    double one_by_one_fill = 0.0;

    if (n != results[num_results - 1].n)
      {
	fprintf (output, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "n = %d used for best result of sparse benchmark is "
		 "different than n = %d used\n"
		 "for worst result of sparse benchmark!\n", 
		 n, results[num_results - 1].n);
	p_summary->failure = 1;
	p_summary->errcode = -1;
	return;
      }
    if (n <= 0)
      {
	fprintf (output, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "n = %d which is non-positive!\n", n);
	p_summary->failure = 1;
	p_summary->errcode = -1;
	return;
      }
    best_fill = ((double) results[0].nnzb * 
		 (double) best_block_size) / n_squared;
    worst_fill = ((double) results[num_results - 1].nnzb * 
		  (double) worst_block_size) / n_squared;
    
    if (results[0].t_median <= 0.0)
      {
	fprintf (output, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with fastest median completion time\n"
		 "finished in non-positive median time!\n");
	p_summary->failure = 1;
	p_summary->errcode = -1;
	return;
      }
    if (results[0].mflops <= 0.0)
      {
	fprintf (output, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with fastest median completion time\n"
		 "involves zero measured floating point operations!\n");
	p_summary->failure = 1;
	p_summary->errcode = -1;
	return;
      }
    best_mflops_rate = results[0].mflops / results[0].t_median;

    if (results[num_results - 1].t_median <= 0.0)
      {
	fprintf (output, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with slowest median completion time\n"
		 "finished in non-positive median time!\n");
	p_summary->failure = 1;
	p_summary->errcode = -1;
	return;
      }
    if (results[num_results - 1].mflops <= 0.0)
      {
	fprintf (output, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Sparse result with slowest median completion time\n"
		 "involves zero measured floating point operations!\n");
	p_summary->failure = 1;
	p_summary->errcode = -1;
	return;
      }
    worst_mflops_rate = results[num_results - 1].mflops / 
      results[num_results - 1].t_median;

    for (i = 0; i < num_results; i++)
      {
	if (results[i].r == 1 && results[i].c == 1)
	  {
	    one_by_one_mflops_rate = results[i].mflops / results[i].t_median;
	    one_by_one_fill = (double) results[i].nnzb / n_squared;
	    break;
	  }
      }
    if (one_by_one_mflops_rate == 0.0)
      {
	fprintf (stderr, "*** ERROR ***\n"
		 "smvm_process_and_output_sparse_results:\n"
		 "Failed to find 1x1 mflops rate!\n");
	die (-1);
      }

    /* Output results */

    p_summary->n = n;

    p_summary->best_r = best_r;
    p_summary->best_c = best_c;
    p_summary->best_block_size = best_block_size;
    p_summary->worst_r = worst_r;
    p_summary->worst_c = worst_c;
    p_summary->worst_block_size = worst_block_size;

    p_summary->best_mflops_rate = best_mflops_rate;
    p_summary->best_fill = best_fill;
    p_summary->worst_mflops_rate = worst_mflops_rate;
    p_summary->worst_fill = worst_fill;

    p_summary->one_by_one_mflops_rate = one_by_one_mflops_rate;
    p_summary->one_by_one_fill = one_by_one_fill;

    p_summary->errcode = 0;
    p_summary->failure = 0;
  }
}
				       

/*****************************************************************************/
void
smvm_invoke_with_results (struct SMVM_invoker_traits* p_traits, 
			  struct SMVM_benchmark_results* p_summary)
{
  const int dims[] = {1,2,3,4,6,8,12};    /* Block dimensions to use. */
  const int n_dims = 7;
  const int num_results = n_dims * n_dims; /* Max number of different possible results */

  struct SMVM_timing_results* sparse_results = NULL;
  int b_dryrun = (p_traits->output == NULL);  /* Dry run if no output stream specified */
  int results_count;
  FILE* out;
  int i, j;

  WITH_DEBUG( fprintf(stderr, ">>> smvm_invoke_with_results <<<\n") );

  /* HPCC wants output to _something_, so at least send to stdout */
  out = p_traits->output;
  if (out == NULL)
    out = stdout;

  sparse_results = 
    smvm_calloc (num_results, sizeof (struct SMVM_timing_results));

  WITH_DEBUG( fprintf(stderr, "With sparse benchmark\n") );

  results_count = 0;
  for (i = 0; i < n_dims; i++)
    for (j = 0; j < n_dims; j++)
      {
	int r = dims[i];
	int c = dims[j];
	
	if (b_dryrun)
	  {
	    WITH_WARN( fprintf (stdout, "Dry run of sparse benchmark with "
				"(r,c) = (%d,%d)\n", r, c) );
	  }
	else
	  {
	    smvm_sparse_single_invocation (p_traits, r, c, 
					   &sparse_results[results_count]);
	    results_count++;
	  }
      }

  smvm_print_verbose_results (p_traits->raw_output_filename, sparse_results, 
			      num_results);
  
  smvm_process_and_store_sparse_results (p_summary, sparse_results, 
					 results_count, out);
}



/*****************************************************************************/
void
smvm_invoke (struct SMVM_invoker_traits* p_traits, int b_run_sparse, 
	     int b_run_dense)
{
  int dims[] = {1,2,3,4,6,8,12};    /* Block dimensions to use. */
  int n_dims = 7;
  int num_results = n_dims * n_dims; /* Max number of different possible results */
  int i, j;
  int results_count;

  WITH_DEBUG( fprintf(stderr, ">>> smvm_invoke:\n") );

  if (b_run_sparse)
    {
      /* Dry run if no output stream specified */
      const int b_dryrun = (p_traits->output == NULL); 
      struct SMVM_timing_results* sparse_results = 
	smvm_calloc (num_results, sizeof (struct SMVM_timing_results));

      WITH_DEBUG( fprintf(stderr, "With sparse benchmark\n") );

      results_count = 0;
      for (i = 0; i < n_dims; i++)
	for (j = 0; j < n_dims; j++)
	  {
	    int r = dims[i];
	    int c = dims[j];

	    if (b_dryrun)
	      {
		WITH_WARN( fprintf (stdout, "Dry run of sparse benchmark with "
				    "(r,c) = (%d,%d)\n", r, c) );
	      }
	    else
	      {
		smvm_sparse_single_invocation (p_traits, r, c, 
					       &sparse_results[results_count]);
		results_count++;
	      }
	  }

      smvm_process_and_output_sparse_results (sparse_results, results_count, 
					      p_traits->output);
    }

  if (b_run_dense)
    {
      /* Dry run if no output stream specified */
      const int b_dryrun = (p_traits->output == NULL); 
      struct SMVM_timing_results* dense_results = 
	smvm_calloc (num_results, sizeof (struct SMVM_timing_results));

      WITH_DEBUG( fprintf(stderr, "With dense benchmark\n") );

      results_count = 0;
      for (i = 0; i < n_dims; i++)
	for (j = 0; j < n_dims; j++)
	  {
	    int r = dims[i];
	    int c = dims[j];

	    if (b_dryrun)
	      {
		WITH_WARN( fprintf (stdout, "Dry run of dense benchmark with "
				    "(r,c) = (%d,%d)\n", r, c) );
	      }
	    else
	      {
		smvm_dense_single_invocation (p_traits, r, c, 
					      &dense_results[results_count]);
		results_count++;
	      }
	  }
    }
}


