#ifndef _smvm_invoker_h
#define _smvm_invoker_h
/**
 * @file smvm_invoker.h
 * @author mfh
 * @date 2004 Nov 21
 *
 * Functions for picking data set dimensions and invoking the benchmark(s).
 *****************************************************************************/
#include "config.h"
#include <stdio.h>

#include "smvm_benchmark_results.h"
#include "smvm_memory_hierarchy_traits.h"
#include "smvm_timing_results.h"


/**
 * @enum SMVM_problem_code
 *
 * Each code describes a different problem that occurred with the invoker.
 * The codes can be OR'd together to get a bit vector.  
 * 
 * @warning  Don't rely on these constants' numeric values!  
 */
enum SMVM_problem_code
{
  /** Indicates no problems. */
  NO_PROBLEMS = 0,  
  /**   
   * Indicates that the data set sizes for the sparse matrix require more 
   * memory than the allowed bounds, so more memory was allocated than was
   * allowed.
   */
  EXCEEDED_MEMORY_BOUND_FOR_SPARSE_MATRIX = 1,
  /**
   * Indicates that the data set sizes for the dense matrix in sparse format
   * require more memory than the allowed bounds, so more memory was 
   * allocated than was allowed.
   */
  EXCEEDED_MEMORY_BOUND_FOR_DENSE_MATRIX_IN_SPARSE_FORMAT = 2,
  /**
   * Some other error.
   */
  OTHER = 4
}; 


/**
 * @struct SMVM_invoker_traits
 *
 * Data for invoking sparse matrix-vector multiplication benchmark.
 */
struct 
SMVM_invoker_traits
{
  /** 
   * Number of times to perform a testing run.  Testing runs should be
   * performed several times, due to the possibility of significant variation
   * in individual runs' results.
   */
  int num_trials;

  /** 
   * Memory hierarchy traits of the host machine.
   */
  struct SMVM_memory_hierarchy_traits memtraits;

  /** Ratio of stored elements to total elements in the sparse matrix. */
  double fill;

  /** Source vector length, and dimension (n x n) of the sparse matrix. */
  int n;   

  /** Number of nonzeros in the sparse matrix. */
  int NNZ; 

  /** Source vector length and dimension (n x n) for dense matrix in sparse format. */
  int dense_n; 

  /**
   * Valid, open output stream in which to store the results of the run.
   * Could be stdout for example.  Client is responsible for managing it
   * (e.g. closing it after use by the invoker).
   */
  FILE* output;

  /**
   * If positive, output status info to stdout, else run silently.  The bigger
   * this is, the more output is shown.  Orthogonal to debuglevel.
   * 
   * @see debuglevel
   */
  int outputlevel;

  /**
   * If positive, output debugging messages to stderr.  The bigger this is,
   * the more output is shown.  Orthogonal to outputlevel.
   * 
   * @see debuglevel
   */
  int debuglevel;

  /**
   * Filename for output of raw (CSV format) benchmark data.  If set to NULL 
   * or "", then no raw output is generated.
   */
  char *raw_output_filename;
};




/**
 * "Constructor" for the SMVM_invoker_traits object (must be pre-allocated).
 *
 * @param fill [IN]  If < 0, uses the default fill of 0.01.
 *
 * @see SMVM_invoker_traits
 */
void
smvm_construct_invocation (struct SMVM_invoker_traits* p_traits, 
			   int num_trials, 
			   double fill,
			   struct SMVM_memory_hierarchy_traits memtraits, 
			   FILE* output, char *raw_output_filename, 
			   int outputlevel, int debuglevel);


/**
 * @brief Computes fill fraction for a sparse matrix.
 *
 * For a sparse matrix, computes the fill resulting from the given count of 
 * nonzeros NNZ, and the dimensions (m,n) of the matrix.
 *
 * @param NNZ  Number of stored ("nonzero") entries in the matrix.
 * @param m    Number of rows in the matrix.
 * @param n    Number of columns in the matrix.
 *
 * @return Fill (real number in [0,1]).
 */
double 
smvm_fill (int NNZ, int m, int n);


/**
 * Returns the number of bytes used by a BSR matrix, where the r x c
 * blocks are of uniform size and are aligned in the matrix.
 *
 * @param NNZ   Number of nonzeros in the matrix.
 * @param m     Number of rows in the matrix.
 * @param n     Number of columns in the matrix.
 * @param r     Number of rows in each register block.
 * @param c     Number of columns in each register block.
 *
 * @return Number of bytes used by the BSR matrix.
 */
int
smvm_bytes_used_by_BSR_matrix (int NNZ, int m, int n, int r, int c);




/**
 * Invokes sparse benchmark for one (r,c) pair.
 *
 * @param p_traits     [in] Contains parameters for the run.
 * @param r            Number of rows in register block. 
 * @param c            Number of columns in register block.
 * @param p_results    Where results are stored.
 *
 * @see smvm_dense_single_invocation()
 */
void
smvm_sparse_single_invocation (struct SMVM_invoker_traits* p_traits, int r, 
			       int c, struct SMVM_timing_results* p_results);


/**
 * Invokes dense-matrix-in-sparse-format benchmark for one (r,c) pair.
 *
 * @param p_traits     [in] Contains parameters for the run.
 * @param r            Number of rows in register block. 
 * @param c            Number of columns in register block.
 * @param p_results    Where results are stored.
 *
 *
 * @see smvm_sparse_single_invocation()
 */
void 
smvm_dense_single_invocation (struct SMVM_invoker_traits* p_traits, int r, 
			       int c, struct SMVM_timing_results* p_results);

/**
 * Run the whole set of BSR SMVM benchmarks, for all the different register
 * block sizes.  
 *
 * @param p_traits        Input parameters for the benchmarks.
 * @param b_run_sparse    If nonzero, run the sparse BSR SMVM benchmark.
 * @param b_run_dense     If nonzero, run the dense-matrix-in-sparse-format 
 *                        BSR SMVM benchmark.
 */
void
smvm_invoke (struct SMVM_invoker_traits* p_traits, int b_run_sparse, 
	     int b_run_dense);

/**
 * Run the sparse BSR sparse-matrix vector multiplication benchmark, for all
 * different register block dimensions; results returned via p_results.  Unlike
 * smvm_invoke(), only runs the sparse benchmark and not the dense in sparse 
 * format benchmark.
 *
 * @param p_traits        Input parameters for the benchmarks.
 * @param p_results       Outputs of the benchmarks.
 */
void
smvm_invoke_with_results (struct SMVM_invoker_traits* p_traits, 
			  struct SMVM_benchmark_results* p_results);

/**
 * Prints _all_ the benchmark results, not just the summary report, to the given
 * output file.  The function smvm_invoke prints out a summary report, but this
 * function prints out everything, in a raw form (comma-delimited ASCII).  If the 
 * file has already been created, it appends to the file; otherwise it creates 
 * the file and first prints column labels before printing the raw data.
 *
 * @return Zero if successful, nonzero if an error occurs.
 */
int
smvm_print_verbose_results (char *filename, 
			    struct SMVM_timing_results results[],
			    const int num_results);


#endif /* _smvm_invoker_h */

